import logging as _logging
import arcgis
from datetime import datetime
from arcgis.features import FeatureSet
from arcgis.mapping import MapImageLayer
from arcgis.geoprocessing import DataFile, LinearUnit, RasterData
from arcgis.geoprocessing import import_toolbox
from arcgis.geoprocessing._support import _execute_gp_tool

_log = _logging.getLogger(__name__)

# _url = "https://logistics.arcgis.com/arcgis/rest/services/World/ClosestFacility/GPServer"
# closest_facility = import_toolbox(gis.properties.helperServices.asyncClosestFacility.url[:-len('/FindClosestFacilities')], gis, True)
_use_async = True


default_incidents = {
    'fields': [{'alias': 'OBJECTID', 'name': 'OBJECTID', 'type': 'esriFieldTypeOID'},
               {'alias': 'Name', 'name': 'Name', 'type': 'esriFieldTypeString', 'length': 128},
               {'alias': 'ID', 'name': 'ID', 'type': 'esriFieldTypeString', 'length': 128},
               {'alias': 'Additional Time', 'name': 'AdditionalTime', 'type': 'esriFieldTypeDouble'},
               {'alias': 'Additional Distance', 'name': 'AdditionalDistance', 'type': 'esriFieldTypeDouble'},
               {'alias': 'Curb Approach', 'name': 'CurbApproach', 'type': 'esriFieldTypeSmallInteger'}],
    'geometryType': 'esriGeometryPoint', 'displayFieldName': '', 'exceededTransferLimit': False,
    'spatialReference': {'latestWkid': 4326, 'wkid': 4326}, 'features': []}

default_facilities = {
                                'fields': [{'alias': 'OBJECTID', 'name': 'OBJECTID', 'type': 'esriFieldTypeOID'},
                                           {'alias': 'Name', 'name': 'Name', 'type': 'esriFieldTypeString',
                                            'length': 128},
                                           {'alias': 'ID', 'name': 'ID', 'type': 'esriFieldTypeString', 'length': 128},
                                           {'alias': 'Additional Time', 'name': 'AdditionalTime',
                                            'type': 'esriFieldTypeDouble'},
                                           {'alias': 'Additional Distance', 'name': 'AdditionalDistance',
                                            'type': 'esriFieldTypeDouble'},
                                           {'alias': 'Curb Approach', 'name': 'CurbApproach',
                                            'type': 'esriFieldTypeSmallInteger'}], 'geometryType': 'esriGeometryPoint',
                                'displayFieldName': '', 'exceededTransferLimit': False,
                                'spatialReference': {'latestWkid': 4326, 'wkid': 4326}, 'features': []}

default_point_barriers = {
                                'fields': [{'alias': 'OBJECTID', 'name': 'OBJECTID', 'type': 'esriFieldTypeOID'},
                                           {'alias': 'Name', 'name': 'Name', 'type': 'esriFieldTypeString',
                                            'length': 128}, {'alias': 'Barrier Type', 'name': 'BarrierType',
                                                             'type': 'esriFieldTypeInteger'},
                                           {'alias': 'Additional Time', 'name': 'Additional_Time',
                                            'type': 'esriFieldTypeDouble'},
                                           {'alias': 'Additional Distance', 'name': 'Additional_Distance',
                                            'type': 'esriFieldTypeDouble'},
                                           {'alias': 'CurbApproach', 'name': 'CurbApproach',
                                            'type': 'esriFieldTypeSmallInteger'}], 'geometryType': 'esriGeometryPoint',
                                'displayFieldName': '', 'exceededTransferLimit': False,
                                'spatialReference': {'latestWkid': 4326, 'wkid': 4326}, 'features': []}

default_line_barriers = {
                                'fields': [{'alias': 'OBJECTID', 'name': 'OBJECTID', 'type': 'esriFieldTypeOID'},
                                           {'alias': 'Name', 'name': 'Name', 'type': 'esriFieldTypeString',
                                            'length': 128}, {'alias': 'SHAPE_Length', 'name': 'SHAPE_Length',
                                                             'type': 'esriFieldTypeDouble'}],
                                'geometryType': 'esriGeometryPolyline', 'displayFieldName': '',
                                'exceededTransferLimit': False, 'spatialReference': {'latestWkid': 4326, 'wkid': 4326},
                                'features': []}

default_polygon_barriers = {
                                'fields': [{'alias': 'OBJECTID', 'name': 'OBJECTID', 'type': 'esriFieldTypeOID'},
                                           {'alias': 'Name', 'name': 'Name', 'type': 'esriFieldTypeString',
                                            'length': 128}, {'alias': 'Barrier Type', 'name': 'BarrierType',
                                                             'type': 'esriFieldTypeInteger'},
                                           {'alias': 'Scaled Time Factor', 'name': 'ScaledTimeFactor',
                                            'type': 'esriFieldTypeDouble'},
                                           {'alias': 'Scaled Distance Factor', 'name': 'ScaledDistanceFactor',
                                            'type': 'esriFieldTypeDouble'},
                                           {'alias': 'SHAPE_Length', 'name': 'SHAPE_Length',
                                            'type': 'esriFieldTypeDouble'},
                                           {'alias': 'SHAPE_Area', 'name': 'SHAPE_Area',
                                            'type': 'esriFieldTypeDouble'}], 'geometryType': 'esriGeometryPolygon',
                                'displayFieldName': '', 'exceededTransferLimit': False,
                                'spatialReference': {'latestWkid': 4326, 'wkid': 4326}, 'features': []}

default_attributes = {
                                'fields': [{'alias': 'ObjectID', 'name': 'OBJECTID', 'type': 'esriFieldTypeOID'},
                                           {'alias': 'AttributeName', 'name': 'AttributeName',
                                            'type': 'esriFieldTypeString', 'length': 255},
                                           {'alias': 'ParameterName', 'name': 'ParameterName',
                                            'type': 'esriFieldTypeString', 'length': 255},
                                           {'alias': 'ParameterValue', 'name': 'ParameterValue',
                                            'type': 'esriFieldTypeString', 'length': 25}], 'features': [{'attributes': {
                                    'OBJECTID': 1, 'AttributeName': 'Any Hazmat Prohibited',
                                    'ParameterValue': 'PROHIBITED', 'ParameterName': 'Restriction Usage'}}, {
                                                                                                            'attributes': {
                                                                                                                'OBJECTID': 2,
                                                                                                                'AttributeName': 'Avoid Carpool Roads',
                                                                                                                'ParameterValue': 'PROHIBITED',
                                                                                                                'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 3,
                                                                                                            'AttributeName': 'Avoid Express Lanes',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 4,
                                                                                                            'AttributeName': 'Avoid Ferries',
                                                                                                            'ParameterValue': 'AVOID_MEDIUM',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 5,
                                                                                                            'AttributeName': 'Avoid Gates',
                                                                                                            'ParameterValue': 'AVOID_MEDIUM',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 6,
                                                                                                            'AttributeName': 'Avoid Limited Access Roads',
                                                                                                            'ParameterValue': 'AVOID_MEDIUM',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 7,
                                                                                                            'AttributeName': 'Avoid Private Roads',
                                                                                                            'ParameterValue': 'AVOID_MEDIUM',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 8,
                                                                                                            'AttributeName': 'Avoid Roads Unsuitable for Pedestrians',
                                                                                                            'ParameterValue': 'AVOID_HIGH',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 9,
                                                                                                            'AttributeName': 'Avoid Stairways',
                                                                                                            'ParameterValue': 'AVOID_HIGH',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 10,
                                                                                                            'AttributeName': 'Avoid Toll Roads',
                                                                                                            'ParameterValue': 'AVOID_MEDIUM',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 11,
                                                                                                            'AttributeName': 'Avoid Toll Roads for Trucks',
                                                                                                            'ParameterValue': 'AVOID_MEDIUM',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 12,
                                                                                                            'AttributeName': 'Avoid Truck Restricted Roads',
                                                                                                            'ParameterValue': 'AVOID_HIGH',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 13,
                                                                                                            'AttributeName': 'Avoid Unpaved Roads',
                                                                                                            'ParameterValue': 'AVOID_HIGH',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 14,
                                                                                                            'AttributeName': 'Axle Count Restriction',
                                                                                                            'ParameterValue': '0',
                                                                                                            'ParameterName': 'Number of Axles'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 15,
                                                                                                            'AttributeName': 'Axle Count Restriction',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 16,
                                                                                                            'AttributeName': 'Driving a Bus',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 17,
                                                                                                            'AttributeName': 'Driving a Delivery Vehicle',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 18,
                                                                                                            'AttributeName': 'Driving a Taxi',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 19,
                                                                                                            'AttributeName': 'Driving a Truck',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 20,
                                                                                                            'AttributeName': 'Driving an Automobile',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 21,
                                                                                                            'AttributeName': 'Driving an Emergency Vehicle',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 22,
                                                                                                            'AttributeName': 'Height Restriction',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 23,
                                                                                                            'AttributeName': 'Height Restriction',
                                                                                                            'ParameterValue': '0',
                                                                                                            'ParameterName': 'Vehicle Height (meters)'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 24,
                                                                                                            'AttributeName': 'Kingpin to Rear Axle Length Restriction',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 25,
                                                                                                            'AttributeName': 'Kingpin to Rear Axle Length Restriction',
                                                                                                            'ParameterValue': '0',
                                                                                                            'ParameterName': 'Vehicle Kingpin to Rear Axle Length (meters)'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 26,
                                                                                                            'AttributeName': 'Length Restriction',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 27,
                                                                                                            'AttributeName': 'Length Restriction',
                                                                                                            'ParameterValue': '0',
                                                                                                            'ParameterName': 'Vehicle Length (meters)'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 28,
                                                                                                            'AttributeName': 'Preferred for Pedestrians',
                                                                                                            'ParameterValue': 'PREFER_LOW',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 29,
                                                                                                            'AttributeName': 'Riding a Motorcycle',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 30,
                                                                                                            'AttributeName': 'Roads Under Construction Prohibited',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 31,
                                                                                                            'AttributeName': 'Semi or Tractor with One or More Trailers Prohibited',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 32,
                                                                                                            'AttributeName': 'Single Axle Vehicles Prohibited',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 33,
                                                                                                            'AttributeName': 'Tandem Axle Vehicles Prohibited',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 34,
                                                                                                            'AttributeName': 'Through Traffic Prohibited',
                                                                                                            'ParameterValue': 'AVOID_HIGH',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 35,
                                                                                                            'AttributeName': 'Truck with Trailers Restriction',
                                                                                                            'ParameterValue': '0',
                                                                                                            'ParameterName': 'Number of Trailers on Truck'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 36,
                                                                                                            'AttributeName': 'Truck with Trailers Restriction',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 37,
                                                                                                            'AttributeName': 'Use Preferred Hazmat Routes',
                                                                                                            'ParameterValue': 'PREFER_MEDIUM',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 38,
                                                                                                            'AttributeName': 'Use Preferred Truck Routes',
                                                                                                            'ParameterValue': 'PREFER_MEDIUM',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 39,
                                                                                                            'AttributeName': 'WalkTime',
                                                                                                            'ParameterValue': '5',
                                                                                                            'ParameterName': 'Walking Speed (km/h)'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 40,
                                                                                                            'AttributeName': 'Walking',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 41,
                                                                                                            'AttributeName': 'Weight Restriction',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 42,
                                                                                                            'AttributeName': 'Weight Restriction',
                                                                                                            'ParameterValue': '0',
                                                                                                            'ParameterName': 'Vehicle Weight (kilograms)'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 43,
                                                                                                            'AttributeName': 'Weight per Axle Restriction',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 44,
                                                                                                            'AttributeName': 'Weight per Axle Restriction',
                                                                                                            'ParameterValue': '0',
                                                                                                            'ParameterName': 'Vehicle Weight per Axle (kilograms)'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 45,
                                                                                                            'AttributeName': 'Width Restriction',
                                                                                                            'ParameterValue': 'PROHIBITED',
                                                                                                            'ParameterName': 'Restriction Usage'}},
                                                                                                        {'attributes': {
                                                                                                            'OBJECTID': 46,
                                                                                                            'AttributeName': 'Width Restriction',
                                                                                                            'ParameterValue': '0',
                                                                                                            'ParameterName': 'Vehicle Width (meters)'}}],
                                'displayFieldName': '', 'exceededTransferLimit': False}

default_tolerance = {'distance': 10, 'units': 'esriMeters'}



def find_closest_facilities(
    incidents,
    facilities,
    measurement_units="Minutes",
    analysis_region=None,
    number_of_facilities_to_find=1,
    cutoff=None,
    travel_direction="Incident to Facility",
    use_hierarchy=True,
    time_of_day=None,
    time_of_day_usage="Start Time",
    uturn_at_junctions="Allowed Only at Intersections and Dead Ends",
    point_barriers=None,
    line_barriers=None,
    polygon_barriers=None,
    restrictions=None,
    attribute_parameter_values=None,
    route_shape="True Shape",
    route_line_simplification_tolerance=None,
    populate_directions=False,
    directions_language ="en",
    directions_distance_units="Miles",
    directions_style_name="NA Desktop",
    time_zone_for_time_of_day="Geographically Local",
    travel_mode="Custom",
    impedance="Drive Time",
    save_output_network_analysis_layer=False,
    overrides=None,
    save_route_data=False,
    time_impedance="TravelTime",
    distance_impedance="Kilometers",
    output_format="Feature Set",
    gis=None,
    future=False):
    """
    Finds one or more facilities that are closest from an incident based on travel time or travel distance and outputs the best routes,
    driving directions between the incidents and the chosen facilities, and a copy of the chosen facilities. You can use the tool, for example,
    to find the closest hospital to an accident, the closest police cars to a crime scene, or the closest store to a customer's address., When
    finding closest facilities, you can specify how many to find and whether the direction of travel is toward or away from them. You can also
    specify the time of day to account for travel times based on live or predictive traffic conditions for that time and date. For instance,
    you can use the tool to search for hospitals within a 15-minute drive time of the site of an accident at a given time of day. Any hospitals
    that take longer than 15 minutes to reach based on the traffic conditions will not be included in the results.

    ======================================  ==========================================================================================================================================
    **Argument**                            **Description**
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    incidents                               Required FeatureSet. Specify one or more incidents (up to 1,000). These are the locations from which the
                                            tool searches for the nearby locations.
                                            When specifying the incidents, you can set properties for each one, such as its name or service time,
                                            by using attributes. The incidents can be specified with the following attributes:

                                                * ``Name`` - The name of the incident. The name is used in the driving directions. If the name is
                                                  not specified, a unique name prefixed with Location is automatically generated in the output routes and directions.
                                                * ``ID`` - A unique identifier for the incident. The identifier is included in the output routes (as
                                                  the IncidentID field) and can help join additional information from the output routes, such as the
                                                  total travel time or total distance, to attributes from your incidents or vice versa. If the ID isn't
                                                  specified, the service autogenerates a unique identifier for each incident.
                                                * ``AdditionalTime``-The amount
                                                  of time spent at the incident, which is added to the total time of the route. The units for this attribute
                                                  value are specified by the Measurement Units parameter. The attribute value is included in the analysis
                                                  only when the measurement units are time based. The default value is 0. If you are finding the closest
                                                  fire stations from fire incidents to estimate response times, the ``AdditionalTime`` attribute can store the
                                                  amount of time it takes firefighters to hook up their equipment at the location of the incident before
                                                  they can begin fighting the fire.
                                                * ``AdditionalDistance``- The extra distance traveled at the incident, which
                                                  is added to the total distance of the route. The units for this attribute value are specified by the
                                                  Measurement Units parameter. The attribute value is included in the analysis only when the measurement
                                                  units are distance based. The default value is 0. Generally, the location of an incident, such as a home,
                                                  isn't exactly on the streets; it is set back somewhat from the road. This attribute value can be used to
                                                  model the distance between the actual incident location and its location on the street, if it is important
                                                  to include that distance in the total travel distance.
                                                * ``CurbApproach`` - Specifies the direction a vehicle may arrive at and depart from the incident.
                                                  The field value is specified as one of the following integers (use the numeric code, not the name in parentheses):

                                                  =========================  ===============================================================
                                                  **Setting**                **Description**
                                                  -------------------------  ---------------------------------------------------------------
                                                  Either side of vehicle     |either|
                                                                              The vehicle can approach and depart the order in either
                                                                              direction, so a U-turn is allowed at the order. This setting
                                                                              can be chosen if it is possible and desirable for your vehicle
                                                                              to turn around at the order. This decision may depend on the
                                                                              width of the road and the amount of traffic or whether the
                                                                              order has a parking lot where vehicles can pull in and turn
                                                                              around.
                                                  -------------------------  ---------------------------------------------------------------
                                                  right side of vehicle      |right|
                                                                              When the vehicle approaches and departs the order, the order must
                                                                              be on the right side of the vehicle. A U-turn is prohibited. This is
                                                                              typically used for vehicles like buses that must arrive with the bus
                                                                              stop on the right side.
                                                  -------------------------  ---------------------------------------------------------------
                                                  left side of vehicle       |left|
                                                                              When the vehicle approaches and departs the order, the curb must
                                                                              be on the left side of the vehicle. A U-turn is prohibited.
                                                                              This is typically used for vehicles like buses that must arrive
                                                                              with the bus stop on the left-hand side.
                                                  -------------------------  ---------------------------------------------------------------
                                                  No U-Turn                  |turn|
                                                                              When the vehicle approaches the order, the curb can be on either side
                                                                              of the vehicle; however, the vehicle must depart without turning around.
                                                  =========================  ===============================================================

                                                  The ``CurbApproach`` property is designed to work with both kinds of national driving standards: right-hand traffic (United States) and
                                                  left-hand traffic (United Kingdom). First, consider an incident on the left side of a vehicle. It is always on the left side regardless
                                                  of whether the vehicle travels on the left or right half of the road. What may change with national driving standards is your decision
                                                  to approach an incident from one of two directions, that is, so it ends up on the right or left side of the vehicle. For example, if you
                                                  want to arrive at an incident and not have a lane of traffic between the vehicle and the incident, you would choose Right side of vehicle (1)
                                                  in the United States but Left side of vehicle (2) in the United Kingdom.
                                                * ``Attr_TravelTime``: Specifies the amount of time for cars, in minutes, that will be added to the total travel time of the route between the
                                                  incident and the closest facility. The attribute value can be used to model the time spent at the incident. For example, if you are finding
                                                  the three closest fire stations from a fire incident, the attribute can store the amount of time spent at the fire incident. This could be
                                                  the time it takes for firefighters to hook up their equipment and begin fighting the fire.
                                                  The value for this attribute is included in the total travel time for the route and is also displayed in driving directions as service time.
                                                  A zero or null value indicates that the incident requires no service time. The default value is 0.
                                                * ``Attr_TruckTravelTime``: Specifies the amount of time for trucks, in minutes, that will be added to the total travel time of the route between
                                                  the incident and the closest facility. The attribute value can be used to model the time spent at the incident.
                                                  The value for this attribute is included in the total travel time for the route and is also displayed in driving directions as service time. A zero or null value indicates
                                                  that the incident requires no service time. The default value is 0.
                                                * ``Attr_WalkTime``: Specifies the amount of time for pedestrians, in minutes, that will be added to the total travel time of the route
                                                  between the incident and the closest facility. The attribute value can be used to model the time spent at the incident.
                                                  The value for this attribute is included in the total travel time for the route and is also displayed in walking directions as service time.
                                                  A zero or null value indicates that the incident requires no service time. The default value is 0.
                                                * ``Attr_Miles``: Specifies the distance in miles that will be added to the total distance of the route between
                                                  the incident and the closest facility. Generally the locations of the incidents are not exactly on the streets
                                                  but are set back somewhat from the road. This attribute value can be used to model the distance between the
                                                  actual incident location and its location on the street if it is important to include that distance in the
                                                  total travel distance. The default value is 0.
                                                * ``Attr_Kilometers``: Specifies the distance in kilometers that will added to the total distance of the route
                                                  between the incident and the closest facility. Generally the locations of the incidents are not exactly on the
                                                  streets but are set back somewhat from the road. This attribute value can be used to model the distance between
                                                  the actual incident location and its location on the street if it is important to include that distance in the
                                                  total travel distance. The default value is 0.
                                                * ``Cutoff_TravelTime``: Specify the travel time for cars, in minutes, at which to stop searching for facilities
                                                  for a given incident. Any incident beyond the cutoff value will not be searched.
                                                  If ``Cutoff_TravelTime`` is not set for an incident, the service will use the value specified as the defa``ultCutoff parameter.
                                                  The value for the ``Cutoff_TravelTime`` attribute allows the ability to overwrite the ``defaultCutoff`` value on a per incident basis.
                                                  The default value for this attribute is null which indicates not to use any cutoff.
                                                * ``Cutoff_TruckTravelTime``: Specify the travel time for trucks, in minutes, at which to stop searching for facilities for a
                                                  given incident. Any incident beyond the cutoff value will not be searched.
                                                  If ``Cutoff_TruckTravelTime`` is not set for an incident, the service will use the value specified as the ``defaultCutoff`` parameter.
                                                  The value for the ``Cutoff_TruckTravelTime`` attribute allows the ability to overwrite the ``defaultCutoff`` value on a per incident basis.
                                                  The default value for this attribute is null which indicates not to use any cutoff.
                                                * ``Cutoff_WalkTime``: Specify the travel time for pedestrians, in minutes, at which to stop searching for facilities for a given incident. Any incident beyond the cutoff value will not be searched.
                                                  If Cutoff_WalkTime is not set for an incident, the service will use the value specified as the ``defaultCutoff`` parameter. The value for the Cutoff_WalkTime attribute allows the ability to overwrite the defaultCutoff value on a per incident basis.
                                                  The default value for this attribute is null which indicates not to use any cutoff.
                                                * ``Cutoff_Miles``: Specify the travel distance in miles at which to stop searching for facilities for a given incident. Any incident beyond the cutoff value will not be searched.
                                                  If ``Cutoff_Miles`` is not set for an incident, the service will use the value specified as the ``defaultCutoff`` parameter. The value for the Cutoff_Miles attribute allows the
                                                  ability to overwrite the ``defaultCutoff`` value on a per incident basis.
                                                  The default value for this attribute is null which indicates not to use any cutoff.
                                                * ``Cutoff_Kilometers``: Specify the travel distance in kilometers at which to stop searching for facilities for a given incident.
                                                  Any incident beyond the cutoff value will not be searched.
                                                  If ``Cutoff_Kilometers`` is not set for an incident, the service will use the value specified as the ``defaultCutoff`` parameter.
                                                  The value for the ``Cutoff_Kilometers`` attribute allows the ability to overwrite the ``defaultCutoff`` value on a per incident basis.
                                                  The default value for this attribute is null which indicates not to use any cutoff.
                                                * ``TargetFacilityCount``: Specify the number of facilities that need to be found for the incident.
                                                  If ``TargetFacilityCount`` is not set for an incident, the service will use the value specified as the
                                                  ``defaultTargetFacilityCount`` parameter. The value for the ``TargetFacilityCount`` attribute allows the ability to
                                                  overwrite the ``defaultTargetFacilityCount`` value on a per incident basis. The default value for this attribute
                                                  is null which causes the service to use the value set for the ``defaultTargetFacilityCount`` parameter. If the
                                                  ``TargetFacilityCount`` attribute is set to a value other than null, the ``defaultTargetFacilityCount`` value is
                                                  overwritten.
                                                * ``Bearing``: Specify the direction the vehicle or person is moving in. ``Bearing`` is measured clockwise from true
                                                  north and must be in degrees. Typically, values are between 0 and 360; however, negative values are interpreted
                                                  by subtracting them from 360 degrees.
                                                * ``BearingTol``: Short for bearing tolerance, this field specifies the maximum acceptable difference between the heading
                                                  of a vehicle and a tangent line from the point on a street where Network Analyst attempts to locate the vehicle.
                                                  The bearing tolerance is used to determine whether the direction in which a vehicle is moving generally aligns with the underlying road.
                                                  If they align within the given tolerance, the vehicle is located on that edge; if not, the next nearest eligible edge is evaluated.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    facilities                              Required FeatureSet. Specify one or more facilities (up to 1,000). These are the locations that are searched for when
                                            finding the closest location. mWhen specifying the facilities, you can set properties for each one, such as its
                                            name or service time, by using attributes. The facilities can be specified with the following attributes:

                                                * ``Name`` - The name of the facility. The name is used in the driving
                                                  directions. If the name is not specified, a unique name prefixed
                                                  with Location is automatically generated in the output routes and
                                                  directions.
                                                * ``ID`` - A unique identifier for the facility. The identifier is included in the output routes (as the
                                                  FacilityID field) and the output closest facilities as FacilityID fields. The FacilityID field can be used
                                                  to join additional information from the output routes, such as the total travel time or total distance,
                                                  to attributes from your facilities. If the ID isn't specified, the service autogenerates a unique identifier
                                                  for each facility.
                                                * ``AdditionalTime``-The amount of time spent at the facility, which is added to the total time
                                                  of the route. The units for this attribute value are specified by the Measurement Units parameter. The
                                                  attribute value is included in the analysis only when the measurement units are time based. The default value is 0.
                                                  If you are finding the closest fire stations to fire incidents, ``AdditionalTime`` can store the
                                                  time it tends to takes a crew to don the appropriate protective equipment
                                                  and exit the fire station.
                                                * ``AdditionalDistance`` - The extra distance traveled at the facility, which is added to the total distance of
                                                  the route. The units for this attribute value are specified by the Measurement Units parameter. The attribute
                                                  value is included in the analysis only when the measurement units are distance based. The default value is 0.
                                                  Generally the location of a facility, such as a fire station, isn't exactly on a street; it is set back somewhat
                                                  from the road. ``AdditionalDistance`` can model the distance between the actual facility location and its location on
                                                  the street, if it is important to include that distance in the total travel distance.
                                                * ``CurbApproach``:  Specifies the direction a vehicle may arrive at and depart
                                                  from the facility. The field value is specified as one of the
                                                  following integers (use the numeric code, not the name in parentheses):

                                                  =========================  ===============================================================
                                                  **Setting**                **Description**
                                                  -------------------------  ---------------------------------------------------------------
                                                  Either side of vehicle     |either|
                                                                             The vehicle can approach and depart the order in either
                                                                             direction, so a U-turn is allowed at the order. This setting
                                                                             can be chosen if it is possible and desirable for your vehicle
                                                                             to turn around at the order. This decision may depend on the
                                                                             width of the road and the amount of traffic or whether the
                                                                             order has a parking lot where vehicles can pull in and turn
                                                                             around.
                                                  -------------------------  ---------------------------------------------------------------
                                                  right side of vehicle      |right|
                                                                             When the vehicle approaches and departs the order, the order must
                                                                             be on the right side of the vehicle. A U-turn is prohibited. This is
                                                                             typically used for vehicles like buses that must arrive with the bus
                                                                             stop on the right side.
                                                  -------------------------  ---------------------------------------------------------------
                                                  left side of vehicle       |left|
                                                                             When the vehicle approaches and departs the order, the curb must
                                                                             be on the left side of the vehicle. A U-turn is prohibited.
                                                                             This is typically used for vehicles like buses that must arrive
                                                                             with the bus stop on the left-hand side.
                                                  -------------------------  ---------------------------------------------------------------
                                                  No U-Turn                  |turn|
                                                                             When the vehicle approaches the order, the curb can be on either side
                                                                             of the vehicle; however, the vehicle must depart without turning around.
                                                  =========================  ===============================================================

                                                  The ``CurbApproach`` property was designed to work with both kinds of national driving standards: right-hand
                                                  traffic (United States) and left-hand traffic (United Kingdom). First, consider a facility on the left side
                                                  of a vehicle. It is always on the left side regardless of whether the vehicle travels on the left or right
                                                  half of the road. What may change with national driving standards is your decision to approach a facility
                                                  from one of two directions, that is, so it ends up on the right or left side of the vehicle. For example,
                                                  if you want to arrive at a facility and not have a lane of traffic between the vehicle and the facility,
                                                  you would choose Right side of vehicle (1) in the United States but Left side of vehicle (2) in the United Kingdom.
                                                * ``Attr_TravelTime``: Specifies the amount of time for cars, in minutes, that will be added to the total travel
                                                  time of the route between the incident and the closest facility. The attribute value can be used to specify the
                                                  turnout time for the facility. For example, when finding the three closest fire stations from a fire incident,
                                                  this attribute can be used to store the time it takes a crew to don the appropriate protective equipment and exit
                                                  the fire station.
                                                  The value for this attribute is included in the total travel time for the route and is also displayed in driving
                                                  directions as service time. A zero or null value indicates that the facility requires no service time. The default
                                                  value is 0.
                                                * ``Attr_TruckTravelTime``: Specifies the amount of time for trucks, in minutes, that will be added to the total
                                                  travel time of the route between the incident and the closest facility.
                                                  The value for this attribute is included in the total travel time for the route and is also displayed in driving
                                                  directions as service time. A zero or null value indicates that the facility requires no service time. The default value is 0.
                                                * ``Attr_WalkTime``: Specifies the amount of time for pedestrians, in minutes, that will be added to the total travel time
                                                  of the route between the incident and the closest facility.
                                                  The value for this attribute is included in the total travel time for the route and is also displayed in walking directions
                                                  as service time. A zero or null value indicates that the facility requires no service time. The default value is 0.
                                                * ``Attr_Miles``: Specifies the distance in miles that will added to the total distance of the route between the incident and
                                                  the closest facility. Generally the locations of the facilities are not exactly on the streets but are set back somewhat from the
                                                  road. This attribute value can be used to model the distance between the actual facility location and its location on the street if
                                                  it is important to include that distance in the total travel distance. The default value is 0.
                                                * ``Attr_Kilometers``: Specifies the distance in kilometers that will added to the total distance of the route between the incident
                                                  and the closest facility. Generally the locations of the facilities are not exactly on the streets but are set back somewhat from the road.
                                                  This attribute value can be used to model the distance between the actual facility location and its location on the street if it is important
                                                  to include that distance in the total travel distance. The default value is 0.
                                                * ``Cutoff_TravelTime``: Specify the travel time for cars, in minutes, at which to stop searching for facilities for a given incident.
                                                  Any incident beyond the cutoff value will not be searched.
                                                  If ``Cutoff_TravelTime`` is not set for a facility, the service will use the value specified as the ``defaultCutoff`` parameter.
                                                  The value for the ``Cutoff_TravelTime`` attribute allows the ability to overwrite the ``defaultCutoff`` value on a per facility basis.
                                                  The default value for this attribute is null which indicates not to use any cutoff.
                                                * ``Cutoff_TruckTravelTime``: Specify the travel time for trucks, in minutes, at which to stop searching for facilities for a given incident. Any incident beyond the cutoff value will not be searched.
                                                  If ``Cutoff_TruckTravelTime`` is not set for a facility, the service will use the value specified as the ``defaultCutoff`` parameter. The value for the Cutoff_TruckTravelTime attribute allows the ability
                                                  to overwrite the ``defaultCutoff`` value on a per facility basis. The default value for this attribute is null which indicates not to use any cutoff.
                                                * ``Cutoff_WalkTime``: Specify the travel time for trucks, in minutes, at which to stop searching for facilities for a given incident. Any incident beyond the cutoff value will not be searched.
                                                  ``Cutoff_WalkTime`` is not set for a facility, the service will use the value specified as the ``defaultCutoff`` parameter. The value for the Cutoff_WalkTime attribute allows the ability to overwrite the defaultCutoff value on a per facility basis. The default value for this attribute is null which indicates not to use any cutoff.
                                                * ``Cutoff_Miles``: Specify the travel distance in miles at which to stop searching for facilities for a given incident. Any incident beyond the cutoff value will not be searched.
                                                  If ``Cutoff_Miles`` is not set for a facility, the service will use the value specified as the ``defaultCutoff`` parameter. The value for the Cutoff_Miles attribute allows the ability to overwrite the defaultCutoff value on a per facility basis. The default value for this attribute is null which indicates not to use any cutoff.
                                                * ``Cutoff_Kilometers``: Specify the travel distance in kilometers at which to stop searching for facilities for a given incident. Any incident beyond the cutoff value will not be searched.
                                                  If Cutoff_Kilometers is not set for a facility, the service will use the value specified as the ``defaultCutoff`` parameter. The value for the Cutoff_Kilometers attribute allows the ability to overwrite the defaultCutoff value on a per facility basis. The default value for this attribute is null which indicates not to use any cutoff.
                                                * ``Bearing``: Specify the direction the vehicle or person is moving in. ``Bearing`` is measured clockwise from true north and must be in degrees.
                                                  Typically, values are between 0 and 360; however, negative values are interpreted by subtracting them from 360 degrees.
                                                * ``BearingTol``: Short for bearing tolerance, this field specifies the maximum acceptable difference between the heading of a vehicle and a tangent
                                                  line from the point on a street where Network Analyst attempts to locate the vehicle. The bearing tolerance is
                                                  used to determine whether the direction in which a vehicle is moving
                                                  generally aligns with the underlying road. If they align within the given tolerance, the vehicle is located on
                                                  that edge; if not, the next nearest eligible edge is evaluated.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    measurement_units                       Required parameter. Specify the units that should be used to measure and report the total travel time or travel
                                            distance for the output routes. The tool finds the closest facility by measuring the travel time or the travel distance
                                            along streets.
                                            The units you choose for this parameter determine whether the tool will measure driving distance or driving time to find
                                            what is closest. Choose a time unit to measure driving time. To measure driving distance, choose a distance unit. Your
                                            choice also determines in which units the tool will report total driving time or distance in the results.

                                            Choice list:['Meters', 'Kilometers', 'Feet', 'Yards', 'Miles', 'Nautical Miles', 'Seconds', 'Minutes', 'Hours', 'Days']
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    analysis_region                         Optional string. Specify the region in which to perform the analysis. If a value is not specified for this parameter, the tool
                                            will automatically calculate the region name based on the location
                                            of the input points. Setting the name of the region is recommended to speed up the
                                            tool execution.

                                            Choice list:['NorthAmerica', 'SouthAmerica', 'Europe', 'MiddleEastAndAfrica', 'India', 'SouthAsia', 'SouthEastAsia', 'Thailand', 'Taiwan', 'Japan', 'Oceania', 'Greece', 'Korea']
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    number_of_facilities_to_find            Optional integer. Specify the number of
                                            closest facilities to find per incident. This is useful in
                                            situations, such as a fire, where multiple fire engines may be
                                            required from different fire stations. You can specify, for
                                            example, to find the three nearest fire stations to a fire.
                                            The tool can find up to 100 facilities from each incident.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    cutoff                                  Optional float. Specify the travel time or travel distance value at which
                                            to stop searching for facilities for a given incident. For
                                            instance, while finding the closest hospitals from the site of an
                                            accident, a cutoff value of 15 minutes would mean that the tool
                                            would search for the closest hospital within 15 minutes from the
                                            incident. If the closest hospital is 17 minutes away, no routes
                                            will be returned in the output routes. A cutoff value is especially
                                            useful when searching for multiple facilities.

                                            The units for this parameter is specified by the Measurement Units parameter.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    travel_direction                        Optional string. Specify whether you want to search for the closest
                                            facility as measured from the incident to the facility or from the facility to the incident.

                                                * Facility to Incident - Direction of travel is from
                                                  facilities to incidents.

                                                * Incident to Facility - Direction of travel is from
                                                  incidents to facilities.

                                            Using one of the parameter values can find different
                                            facilities as the travel time along some streets may vary based on
                                            the travel direction and one-way restrictions. For instance, a
                                            facility may be a 10-minute drive from the incident while traveling
                                            from the incident to the facility, but while traveling from the
                                            facility to the incident, it may be a 15-minute journey because of
                                            different travel time in that direction. If you are setting a value for Time of Day, traffic may also cause the Facility to Incident and Incident to Facility options to return different results.

                                            Fire departments commonly use the Facility to Incident
                                            value for the parameter since they are concerned with the time it
                                            takes to travel from the fire station (facility) to the location of
                                            the emergency (incident). A retail store (facility) is more
                                            concerned with the time it takes the shoppers (incidents) to reach
                                            the store; therefore, stores commonly choose Incident to Facility.

                                            Choice list: ['Incident to Facility', 'Facility to Incident']
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    use_hierarchy                           Optional boolean. Specify whether hierarchy should be used when finding the best
                                            route between the facility and the incident.

                                              * Checked (True) - Use hierarchy when finding routes. When
                                                hierarchy is used, the tool prefers higher-order streets (such as
                                                freeways) to lower-order streets (such as local roads), and can be used
                                                to simulate the driver preference of traveling on freeways instead
                                                of local roads even if that means a longer trip. This is especially
                                                true when finding routes to faraway facilities, because drivers on long-distance
                                                trips tend to prefer traveling on freeways where stops, intersections, and turns can be avoided.
                                                Using hierarchy is computationally faster,
                                                especially for long-distance routes, as the tool has to select the
                                                best route from a relatively smaller subset of streets.

                                              * Unchecked (False) - Do not use hierarchy when finding routes. If
                                                hierarchy is not used, the tool considers all the streets and doesn't
                                                prefer higher-order streets when finding the route. This is often
                                                used when finding short-distance routes within a city.

                                            The tool automatically reverts to using hierarchy if the
                                            straight-line distance between facilities and incidents is
                                            greater than 50 miles, even if you have set this parameter to not use hierarchy.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    time_of_day                             Optional datetime. Specifies the time and date at which the routes should
                                            begin or end. The value is used as the start time or end time for
                                            the route depending on the value for the Time of Day Usage
                                            parameter. If you specify the current date and time as the value
                                            for this parameter, the tool will use live traffic conditions to
                                            find the closest facilities and the total travel time will be based
                                            on traffic conditions.

                                            Specifying a time of day results in more accurate
                                            estimation of travel time between the incident and facility because the
                                            travel times account for the traffic conditions that are applicable
                                            for that date and time.
                                            The Time Zone for Time of Day parameter specifies whether this time and date
                                            refer to UTC or the time zone in which the facility or incident is located.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    time_of_day_usage                       Optional string. Indicates whether the Time of Day parameter value
                                            represents the arrival or departure time for the routes.

                                                * Start Time - When this option is chosen, the tool finds
                                                  the best route considering the Time of Day parameter value as the
                                                  departure time from the facility or incident.

                                                * End Time - When this option is chosen, the tool considers
                                                  the Time of Day parameter value as the arrival time at the facility
                                                  or incident. This option is useful if you want to know what time to
                                                  depart from a location so you arrive at the destination at the
                                                  time specified in Time of Day.

                                            Choice list: ['Start Time', 'End Time']
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    uturn_at_junctions                      Optional string. The U-Turn policy at junctions. Allowing U-turns implies the solver can turn around at a junction and double back on the same street.

                                            Given that junctions represent street intersections and dead ends, different vehicles  may be able to turn around
                                            at some junctions but not at others-it depends on whether the junction represents an intersection or dead end.
                                            To accommodate, the U-turn policy parameter is implicitly specified by how many edges connect to the junction,
                                            which is known as junction valency. The acceptable values for this parameter are listed below; each is followed
                                            by a description of its meaning in terms of junction valency.
                                            Allowed-U-turns are permitted at junctions with any number of connected edges. This is the default value.
                                            Not Allowed-U-turns are prohibited at all junctions, regardless of junction valency. Note, however, that
                                            U-turns are still permitted at network locations even when this setting is chosen; however, you can set the
                                            individual network locations' CurbApproach property to prohibit U-turns there as well.  Allowed only at Dead
                                            Ends-U-turns are prohibited at all junctions, except those that have only one adjacent edge (a dead end).
                                            Allowed only at Intersections and Dead Ends-U-turns are prohibited at junctions where exactly two adjacent edges meet
                                            but are permitted at intersections (junctions with three or more adjacent edges) and dead ends (junctions with exactly
                                            one adjacent edge).  Oftentimes, networks have extraneous junctions in the middle of road segments. This option
                                            prevents vehicles from making U-turns at these locations.

                                            Choice list: ['Allowed', 'Not Allowed', 'Allowed Only at Dead Ends', 'Allowed Only at Intersections and Dead Ends']
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    point_barriers                          Optional FeatureSet. Specify one or more points to act as temporary
                                            restrictions or represent additional time or distance that may be
                                            required to travel on the underlying streets. For example, a point
                                            barrier can be used to represent a fallen tree along a street or
                                            time delay spent at a railroad crossing.

                                            The tool imposes a limit of 250 points that can be added
                                            as barriers.
                                            When specifying the point barriers, you can set properties for each one, such as its name or barrier type,
                                            by using attributes. The point barriers can be specified with the following attributes:

                                                * ``Name``: The name of the barrier.
                                                * ``BarrierType``: Specifies whether the point barrier restricts travel
                                                  completely or adds time or distance when it is crossed. The value
                                                  for this attribute is specified as one of the following
                                                  integers (use the numeric code, not the name in parentheses):

                                                    * 0 (Restriction) - Prohibits travel through the barrier. The barrier
                                                    is referred to as a restriction point barrier since it acts as a
                                                    restriction.

                                                    * 2 (Added Cost) - Traveling through the barrier increases the travel
                                                    time or distance by the amount specified in the
                                                    Additional_Time or Additional_Distance field. This barrier type is
                                                    referred to as an added-cost point barrier.
                                                * ``Additional_Time``: Indicates how much travel time is added when the
                                                  barrier is traversed. This field is applicable only for added-cost
                                                  barriers and only if the measurement units are time based. This field
                                                  value must be greater than or equal to zero, and its units are the same as those specified in the
                                                  Measurement Units parameter.
                                                * ``Additional_Distance``: Indicates how much distance is added when the barrier is
                                                  traversed. This field is applicable only for added-cost barriers
                                                  and only if the measurement units are distance based. The field value
                                                  must be greater than or equal to zero, and its units are the same as those specified in the
                                                  Measurement Units parameter.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    line_barriers                           Optional FeatureSet.  Specify one or more lines that prohibit travel anywhere
                                            the lines intersect the streets. For example, a parade or protest
                                            that blocks traffic across several street segments can be modeled
                                            with a line barrier. A line barrier can also quickly fence off
                                            several roads from being traversed, thereby channeling possible
                                            routes away from undesirable parts of the street
                                            network.

                                            The tool imposes a limit on the number of streets you can
                                            restrict using the Line Barriers parameter. While there is no limit on
                                            the number of lines you can specify as line barriers, the combined
                                            number of streets intersected by all the lines cannot exceed
                                            500.
                                            When specifying the line barriers, you can set a name property for each one by using the following attribute:

                                              * ``Name``: The name of the barrier.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    polygon_barriers                        Optional FeatureSet. Specify polygons that either completely restrict travel or
                                            proportionately scale the time or distance required to travel on
                                            the streets intersected by the polygons.

                                            The service imposes a limit on the number of streets you
                                            can restrict using the Polygon Barriers parameter. While there is
                                            no limit on the number of polygons you can specify as the polygon
                                            barriers, the combined number of streets intersected by all the
                                            polygons should not exceed 2,000.
                                            When specifying the polygon barriers, you can set properties for each one, such as its name or barrier type,
                                            by using attributes. The polygon barriers can be specified with the following attributes:

                                                * ``Name``: The name of the barrier.
                                                * ``BarrierType``: Specifies whether the barrier restricts travel completely
                                                  or scales the time or distance for traveling through it. The field
                                                  value is specified as one of the following integers (use the numeric code, not the name in parentheses):

                                                    * 0 (Restriction) - Prohibits traveling through any part of the barrier.
                                                      The barrier is referred to as a restriction polygon barrier since it
                                                      prohibits traveling on streets intersected by the barrier. One use
                                                      of this type of barrier is to model floods covering areas of the
                                                      street that make traveling on those streets impossible.

                                                    * 1 (Scaled Cost) - Scales the time or distance required to travel the
                                                      underlying streets by a factor specified using the ScaledTimeFactor
                                                      or ScaledDistanceFactor fields. If the streets are partially
                                                      covered by the barrier, the travel time or distance is apportioned
                                                      and then scaled. For example, a factor 0.25 would mean that travel
                                                      on underlying streets is expected to be four times faster than
                                                      normal. A factor of 3.0 would mean it is expected to take three
                                                      times longer than normal to travel on underlying streets. This
                                                      barrier type is referred to as a scaled-cost polygon barrier. It
                                                      might be used to model storms that reduce travel speeds in specific
                                                      regions.
                                                * ``ScaledTimeFactor``: This is the factor by which the travel time of the streets
                                                  intersected by the barrier is multiplied. This field is applicable
                                                  only for scaled-cost barriers and only if the measurement units are time
                                                  based. The field value must be greater than zero.
                                                * ``ScaledDistanceFactor``: This is the factor by which the distance of the streets
                                                  intersected by the barrier is multiplied. This attribute is
                                                  applicable only for scaled-cost barriers and only if the measurement
                                                  units are distance based. The attribute value must be greater than
                                                  zero.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    restrictions                            Optional string. Specify which restrictions should be honored by the tool when finding the best routes between
                                            incidents and facilities. A restriction represents a driving preference or requirement. In most cases, restrictions
                                            cause roads to be prohibited, but they can also cause them to be avoided or preferred. For instance, using an Avoid
                                            Toll Roads restriction will result in a route that will include toll roads only when it is absolutely required to
                                            travel on toll roads in order to visit an incident or a facility. Height Restriction makes it possible to route
                                            around any clearances that are lower than the height of your vehicle. If you are carrying corrosive materials on
                                            your vehicle, using the Any Hazmat Prohibited restriction prevents hauling the materials along roads where it is
                                            marked as illegal to do so. Below is a list of available restrictions and a short description.
                                            Some restrictions require an additional value to be
                                            specified for their desired use. This value needs to be associated
                                            with the restriction name and a specific parameter intended to work
                                            with the restriction. You can identify such restrictions if their
                                            names appear under the AttributeName column in the Attribute
                                            Parameter Values parameter. The ParameterValue field should be
                                            specified in the Attribute Parameter Values parameter for the
                                            restriction to be correctly used when finding traversable roads.
                                            Some restrictions are supported only in certain countries; their availability is stated by region in the list below.
                                            Of the restrictions that have limited availability within a region, you can check whether the restriction is available
                                            in a particular country by looking at the table in the Country List section of the Data coverage for network analysis
                                            services web page. If a country has a value of  Yes in the Logistics Attribute column, the restriction with select
                                            availability in the region is supported in that country. If you specify restriction names that are not available in the
                                            country where your incidents are located, the service ignores the invalid restrictions. The service also ignores
                                            restrictions whose Restriction Usage parameter value is between 0 and 1 (see the Attribute Parameter Value parameter).
                                            It prohibits all restrictions whose Restriction Usage parameter value is greater than 0.
                                            The tool supports the following restrictions:

                                            ========================================  ================================================
                                            **Parameter**                             **Description**
                                            ----------------------------------------  ------------------------------------------------
                                            Any Hazmat Prohibited                     The results will not include roads
                                                                                      where transporting any kind of hazardous material is
                                                                                      prohibited.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Carpool Roads                       The results will avoid roads that are
                                                                                      designated exclusively for carpool (high-occupancy)
                                                                                      vehicles.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Express Lanes                       The results will avoid roads designated
                                                                                      as express lanes.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Ferries                             The results will avoid ferries.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Gates                               The results will avoid roads where there are
                                                                                      gates such as keyed access or guard-controlled
                                                                                      entryways.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Limited Access Roads                The results will avoid roads
                                                                                      that are limited access highways.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Private Roads                       The results will avoid roads that are
                                                                                      not publicly owned and maintained.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Roads Unsuitable for Pedestrians    The result will avoid roads that are unsuitable for pedestrians.
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Stairways                           The result will avoid all stairways on a pedestrian suitable route.
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Toll Roads                          The results will avoid toll
                                                                                      roads.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Toll Roads for Trucks               The result will avoid all toll roads for trucks
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Truck Restricted Roads              The result will avoid roads where trucks are not allowed except when making deliveries.
                                            ----------------------------------------  ------------------------------------------------
                                            Avoid Unpaved Roads                       The results will avoid roads that are
                                                                                      not paved (for example, dirt, gravel, and so on).
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Axle Count Restriction                    The results will not include roads
                                                                                      where trucks with the specified number of axles are prohibited. The
                                                                                      number of axles can be specified using the Number of Axles
                                                                                      restriction parameter.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Driving a Bus                             The results will not include roads where
                                                                                      buses are prohibited. Using this restriction will also ensure that
                                                                                      the results will honor one-way streets.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Driving a Taxi                            The results will not include roads where
                                                                                      taxis are prohibited. Using this restriction will also ensure that
                                                                                      the results will honor one-way streets.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Driving a Truck                           The results will not include roads where
                                                                                      trucks are prohibited. Using this restriction will also ensure that
                                                                                      the results will honor one-way streets.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Driving an Automobile                     The results will not include roads
                                                                                      where automobiles are prohibited. Using this restriction will also
                                                                                      ensure that the results will honor one-way streets.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Driving an Emergency Vehicle              The results will not include
                                                                                      roads where emergency vehicles are prohibited. Using this
                                                                                      restriction will also ensure that the results will honor one-way
                                                                                      streets.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Height Restriction                        The results will not include roads
                                                                                      where the vehicle height exceeds the maximum allowed height for the
                                                                                      road. The vehicle height can be specified using the Vehicle Height
                                                                                      (meters) restriction parameter.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Kingpin to Rear Axle                      The results will
                                            Length Restriction                        not include roads where the vehicle length exceeds the maximum
                                                                                      allowed kingpin to rear axle for all trucks on the road. The length
                                                                                      between the vehicle kingpin and the rear axle can be specified
                                                                                      using the Vehicle Kingpin to Rear Axle Length (meters) restriction
                                                                                      parameter.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Length Restriction                        The results will not include roads
                                                                                      where the vehicle length exceeds the maximum allowed length for the
                                                                                      road. The vehicle length can be specified using the Vehicle Length
                                                                                      (meters) restriction parameter.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Preferred for Pedestrians                 The result prefers paths designated for pedestrians.
                                            ----------------------------------------  ------------------------------------------------
                                            Riding a Motorcycle                       The results will not include roads
                                                                                      where motorcycles are prohibited. Using this restriction will also
                                                                                      ensure that the results will honor one-way streets.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Roads Under Construction Prohibited       The results will not
                                                                                      include roads that are under construction.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Semi or Tractor with One                  The results will not include roads where semis or tractors with
                                            or More Trailers Prohibited               one or more trailers are prohibited.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Single Axle Vehicles Prohibited           The results will not
                                                                                      include roads where vehicles with single axles are
                                                                                      prohibited.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Tandem Axle Vehicles Prohibited           The results will not
                                                                                      include roads where vehicles with tandem axles are
                                                                                      prohibited.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Through Traffic Prohibited                The results will not include
                                                                                      roads where through traffic (non local) is prohibited.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Truck with Trailers Restriction           The results will not
                                                                                      include roads where trucks with the specified number of trailers on
                                                                                      the truck are prohibited. The number of trailers on the truck can
                                                                                      be specified using the Number of Trailers on Truck restriction
                                                                                      parameter.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Use Preferred Hazmat Routes               The results will prefer roads
                                                                                      that are designated for transporting any kind of hazardous
                                                                                      materials.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Use Preferred Truck Routes                The results will prefer roads
                                                                                      that are designated as truck routes, such as the roads that are
                                                                                      part of the national network as specified by the National Surface
                                                                                      Transportation Assistance Act in the United States, or roads that
                                                                                      are designated as truck routes by the state or province, or roads
                                                                                      that are preferred by the trucks when driving in an
                                                                                      area.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Walking                                   The results will not include roads where
                                                                                      pedestrians are prohibited.
                                                                                      Availability: All countries
                                            ----------------------------------------  ------------------------------------------------
                                            Weight Restriction                        The results will not include roads
                                                                                      where the vehicle weight exceeds the maximum allowed weight for the
                                                                                      road. The vehicle weight can be specified using the Vehicle Weight
                                                                                      (kilograms) restriction parameter.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Weight per Axle Restriction               The results will not include
                                                                                      roads where the vehicle weight per axle exceeds the maximum allowed
                                                                                      weight per axle for the road. The vehicle weight per axle can be
                                                                                      specified using the Vehicle Weight per Axle (kilograms) restriction
                                                                                      parameter.
                                                                                      Availability: Select countries in North America and Europe
                                            ----------------------------------------  ------------------------------------------------
                                            Width Restriction                         The results will not include roads where
                                                                                      the vehicle width exceeds the maximum allowed width for the road.
                                                                                      The vehicle width can be specified using the Vehicle Width (meters)
                                                                                      restriction parameter.
                                                                                      Availability: Select countries in North America and Europe
                                            ========================================  ================================================

                                            Choice list:['Any Hazmat Prohibited', 'Avoid Carpool Roads', 'Avoid Express Lanes', 'Avoid Ferries', 'Avoid Gates',
                                            'Avoid Limited Access Roads', 'Avoid Private Roads', 'Avoid Roads Unsuitable for Pedestrians', 'Avoid Stairways',
                                            'Avoid Toll Roads', 'Avoid Toll Roads for Trucks', 'Avoid Truck Restricted Roads', 'Avoid Unpaved Roads', 'Axle Count Restriction',
                                            'Driving a Bus', 'Driving a Delivery Vehicle', 'Driving a Taxi', 'Driving a Truck', 'Driving an Automobile', 'Driving an Emergency Vehicle',
                                            'Height Restriction', 'Kingpin to Rear Axle Length Restriction', 'Length Restriction', 'Preferred for Pedestrians', 'Riding a Motorcycle',
                                            'Roads Under Construction Prohibited', 'Semi or Tractor with One or More Trailers Prohibited', 'Single Axle Vehicles Prohibited',
                                            'Tandem Axle Vehicles Prohibited', 'Through Traffic Prohibited', 'Truck with Trailers Restriction', 'Use Preferred Hazmat Routes',
                                            'Use Preferred Truck Routes', 'Walking', 'Weight Restriction', 'Weight per Axle Restriction', 'Width Restriction']
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    attribute_parameter_values              Optional FeatureSet. Specify additional values required by some restrictions, such as the weight of a vehicle for
                                            Weight Restriction. You can also use the attribute parameter to specify whether any restriction prohibits, avoids,
                                            or prefers travel on roads that use the restriction. If the restriction is
                                            meant to avoid or prefer roads, you can further specify the degree
                                            to which they are avoided or preferred using this
                                            parameter. For example, you can choose to never use toll roads, avoid them as much as possible, or even highly prefer them.
                                            The values you provide for this parameter are ignored unless Travel Mode is set to Custom.
                                            If you specify the Attribute Parameter Values parameter from a
                                            feature class, the field names on the feature class must match the fields as described below:

                                              * ``AttributeName``: Lists the name of the restriction.
                                              * ``ParameterName``: Lists the name of the parameter associated with the
                                                restriction. A restriction can have one or more ParameterName field
                                                values based on its intended use.
                                              * ``ParameterValue``: The value for ParameterName used by the tool
                                                when evaluating the restriction.

                                                Attribute Parameter Values is dependent on the
                                                Restrictions parameter. The ParameterValue field is applicable only
                                                if the restriction name is specified as the value for the
                                                Restrictions parameter.

                                                In Attribute Parameter Values, each
                                                restriction (listed as AttributeName) has a ParameterName field
                                                value, Restriction Usage, that specifies whether the restriction
                                                prohibits, avoids, or prefers travel on the roads associated with
                                                the restriction and the degree to which the roads are avoided or
                                                preferred. The Restriction Usage ParameterName can be assigned any of
                                                the following string values or their equivalent numeric values
                                                listed within the parentheses:

                                                    PROHIBITED (-1)-Travel on the roads using the restriction is completely
                                                    prohibited.

                                                    AVOID_HIGH (5) - It
                                                    is highly unlikely for the tool to include in the route the roads
                                                    that are associated with the restriction.

                                                    AVOID_MEDIUM (2) - It
                                                    is unlikely for the tool to include in the route the roads that are
                                                    associated with the restriction.

                                                    AVOID_LOW (1.3) - It
                                                    is somewhat unlikely for the tool to include in the route the roads
                                                    that are associated with the restriction.

                                                    PREFER_LOW (0.8)  -It
                                                    is somewhat likely for the tool to include in the route the roads
                                                    that are associated with the restriction.

                                                    PREFER_MEDIUM (0.5) - It is likely for the tool to include in the route the roads that
                                                    are associated with the restriction.

                                                    PREFER_HIGH (0.2) - It is highly likely for the tool to include in the route the roads
                                                    that are associated with the restriction.

                                                In most cases, you can use the default value, PROHIBITED,
                                                for the Restriction Usage if the restriction is dependent on a
                                                vehicle-characteristic such as vehicle height. However, in some
                                                cases, the value for Restriction Usage depends on your routing
                                                preferences. For example, the Avoid Toll Roads restriction has the
                                                default value of AVOID_MEDIUM for the Restriction Usage parameter.
                                                This means that when the restriction is used, the tool will try to
                                                route around toll roads when it can. AVOID_MEDIUM also indicates
                                                how important it is to avoid toll roads when finding the best
                                                route; it has a medium priority. Choosing AVOID_LOW would put lower
                                                importance on avoiding tolls; choosing AVOID_HIGH instead would
                                                give it a higher importance and thus make it more acceptable for
                                                the service to generate longer routes to avoid tolls. Choosing
                                                PROHIBITED would entirely disallow travel on toll roads, making it
                                                impossible for a route to travel on any portion of a toll road.
                                                Keep in mind that avoiding or prohibiting toll roads, and thus
                                                avoiding toll payments, is the objective for some; in contrast,
                                                others prefer to drive on toll roads because avoiding traffic is
                                                more valuable to them than the money spent on tolls. In the latter
                                                case, you would choose PREFER_LOW, PREFER_MEDIUM, or PREFER_HIGH as
                                                the value for Restriction Usage. The higher the preference, the
                                                farther the tool will go out of its way to travel on the roads
                                                associated with the restriction.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    route_shape                             Optional string. Specify the type of route features that are output by the
                                            tool. The parameter can be specified using one of the following
                                            values:

                                              * True Shape - Return the exact shape of the resulting route
                                                that is based on the underlying streets.

                                              * Straight Line - Return a straight line between the
                                                incident and the facility.

                                              * None - Do not return any shapes for the routes. This value
                                                can be useful in cases where you are only interested in determining
                                                the total travel time or travel distance between the closest
                                                facility and the incident.

                                            When the Route Shape parameter is set to True Shape, the
                                            generalization of the route shape can be further controlled using
                                            the appropriate values for the Route Line Simplification Tolerance
                                            parameters.

                                            No matter which value you choose for the Route Shape
                                            parameter, the best route is always determined by minimizing the
                                            travel time or the travel distance, never using the straight-line
                                            distance between incidents and
                                            facilities. This means that only the route shapes are different,
                                            not the underlying streets that are searched when finding the
                                            route.

                                            Choice list:['True Shape', 'Straight Line', 'None']
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    route_line_simplification_tolerance     Optional LinearUnit. Specify by how much you want to simplify the geometry of the output lines for
                                            routes and directions. The tool ignores this parameter if the Route Shape parameter isn't set to True Shape.
                                            Simplification maintains critical
                                            points on a route, such as turns at intersections, to define the
                                            essential shape of the route and removes other points. The
                                            simplification distance you specify is the maximum allowable offset
                                            that the simplified line can deviate from the original line.
                                            Simplifying a line reduces the number of vertices that are part of
                                            the route geometry. This improves the tool execution
                                            time.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    populate_directions                     Optional boolean. Specify whether the tool should generate driving directions for
                                            each route.
                                            Checked (True): Indicates that the directions will be generated
                                            and configured based on the values for the Directions Language,
                                            Directions Style Name, and Directions Distance Units
                                            parameters.
                                            Unchecked (False): Directions are not generated, and the tool
                                            returns an empty Directions layer.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    directions_language                     Optional string. Specify the language that should be used when generating
                                            driving directions.

                                            This parameter is used only when the Populate
                                            Directions parameter is checked, or set to True.

                                            The parameter value can be
                                            specified using one of the following two- or five-character language codes:

                                            Choice list: ['ar-Arabic', 'cs-Czech', 'de-German', 'el-Greek', 'en-English', 'es-Spanish', 'et-Estonian', 'fr-French', 'he-Hebrew', 'it-Italian',
                                            'ja-Japanese', 'ko-Korean', 'lt-Lithuanian', 'lv-Latvian', 'nl-Dutch', 'pl-Polish', 'pt-BR-Brazilian', 'Portuguese', 'pt-PT-European', 'Portuguese', 'ru-Russian',
                                            'sv-Swedish', 'th-Thai', 'tr-Turkish', 'zh-CN-Simplified', 'Chinese']

                                            If an unsupported language code is specified, the tool
                                            returns the directions using the default language,
                                            English.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    directions_distance_units               Optional string. Specify the units for displaying travel distance in the
                                            driving directions. This parameter is used only when the Populate
                                            Directions parameter is checked, or set to True.
                                            Miles Kilometers Meters Feet Yards NauticalMiles

                                            Choice list: ['Miles', 'Kilometers', 'Meters', 'Feet', 'Yards', 'NauticalMiles']
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    directions_style_name                   Optional string. Specify the name of the formatting style for the
                                            directions. This parameter is used only when the Populate
                                            Directions parameter is checked, or set to True. The parameter can be specified
                                            using the following values:

                                            ``NA Desktop``: Generates turn-by-turn directions suitable for printing.
                                            ``NA Navigation``: Generates turn-by-turn directions designed for an in-vehicle navigation device.

                                            Choice list: ['NA Desktop', 'NA Navigation']
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    time_zone_for_time_of_day               Optional string. Specifies the time zone of the Time of Day parameter.
                                            Geographically Local: The Time of Day parameter refers to the time zone in which the facilities or incidents are
                                            located.  If Time of Day Usage is set to Start Time and Travel Direction is Facility to Incident, this is the time
                                            zone of the facilities.  If Time of Day Usage is set to Start Time and Travel Direction is Incident to Facility,
                                            this is the time zone of the incidents.  If Time of Day Usage is set to End Time and Travel Direction is Facility
                                            to Incident, this is the time zone of the incidents.  If Time of Day Usage is set to End Time and Travel Direction
                                            is Incident to Facility, this is the time zone of the facilities.  UTC: The Time of Day parameter refers to
                                            Coordinated Universal Time (UTC). Choose this option if you want to find what's nearest for a specific time, such as
                                            now, but aren't certain in which time zone the facilities or incidents will be located.
                                            Irrespective of the Time Zone for Time of Day setting, if your facilities
                                            and incidents are in multiple time zones, the following rules are
                                            enforced by the tool:

                                            All incidents must be in the same time zone
                                              when: Specifying a start time and traveling from incident to facility. Specifying an end time and traveling from facility to incident.

                                            All facilities must be in the same time zone
                                              when: Specifying a start time and traveling from facility to incident. Specifying an end time and traveling from incident to facility.

                                            Choice list:['Geographically Local', 'UTC']
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    travel_mode                             Optional string. Specify the mode of transportation to model in the analysis. Travel modes are managed in ArcGIS
                                            Online and can be configured by the administrator of your organization to better reflect your organization's
                                            workflows. You need to specify the name of a travel mode supported by your organization.

                                            To get a list of supported travel mode names, run the GetTravelModes tool from the Utilities toolbox available under the same GIS Server connection
                                            you used to access the tool. The GetTravelModes tool adds a table, Supported Travel Modes, to the application. Any value in the Travel Mode Name field from the
                                            Supported Travel Modes table can be specified as input. You can also specify the value from Travel Mode Settings field as input. This speeds up the tool execution as the
                                            tool does not have to lookup the settings based on the travel mode name.

                                            The default value, Custom, allows you to configure your own travel mode using the custom travel mode parameters (UTurn at Junctions, Use Hierarchy, Restrictions, Attribute Parameter Values, and Impedance).
                                            The default values of the custom travel mode parameters model travelling by car. You may want to choose Custom and set the custom travel mode parameters listed above to model a pedestrian with a fast walking speed
                                            or a truck with a given height, weight, and cargo of certain hazardous materials. You may choose to do this to try out different settings to get desired analysis results.
                                            Once you have identified the analysis settings, you should work with your organization's administrator and save these settings as part of new or existing travel mode so that
                                            everyone in your organization can rerun the analysis with the same settings.
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    impedance                               Optional string. Specify the impedance, which is a value that represents the effort or cost of traveling along
                                            road segments or on other parts of the transportation network.
                                            Travel distance is an impedance; the length of a road in kilometers can be thought of as impedance.
                                            Travel distance in this sense is the same for all modes-a kilometer for a pedestrian is also a kilometer for a car.
                                            (What may change is the pathways on which the different modes are allowed to travel, which affects distance between points,
                                            and this is modeled by travel mode settings.) Travel time can also be an impedance; a car may take one minute to travel a mile
                                            along an empty road. Travel times can vary by travel mode-a pedestrian may take more than 20  minutes to walk the same mile,
                                            so it is important to choose the right impedance for the travel mode you are modeling.  Choose from the following impedance values:
                                            Drive Time-Models travel times for a car. These travel times are dynamic and fluctuate according to traffic flows in areas where
                                            traffic data is available. This is the default value. Truck Time-Models travel times for a truck.  These travel times are static for
                                            each road and don't fluctuate with traffic. Walk Time-Models travel times for a pedestrian. Travel Distance-Stores  length measurements
                                            along roads and paths. To model walk distance, choose this option and ensure Walking is  set in the Restriction parameter. Similarly,
                                            to model drive or truck distance, choose Travel Distance here and set the appropriate restrictions so your vehicle travels only on
                                            roads where it is permitted to do so. The value you provide for this parameter is ignored unless Travel Mode is set to Custom, which
                                            is the default value. If you choose Drive Time, Truck Time, or Walk Time, the Measurement Units parameter must be set to a time-based
                                            value; if you choose Travel Distance for Impedance, Measurement Units must be distance-based.

                                            Choice list:['Drive Time', 'Truck Time', 'Walk Time', 'Travel Distance']
    --------------------------------------  ------------------------------------------------------------------------------------------------------------------------------------------
    gis                                     Optional, the GIS on which this tool runs. If not specified, the active GIS is used.
    ======================================  ==========================================================================================================================================

    ..:: Returns the following as a named tuple:

            * output_routes - Output Routes as a FeatureSet
            * output_directions - Output Directions as a FeatureSet
            * solve_succeeded - Solve Succeeded as a bool
            * output_closest_facilities - Output Closest Facilities as a FeatureSet


    Click `FindClosestFacilities`_ for additional help.
    """
    if gis is None:
        gis = arcgis.env.active_gis
    url = gis.properties.helperServices.asyncClosestFacility.url[:-len('/FindClosestFacilities')]
    tbx = import_toolbox(url, gis=gis)
    defaults = dict(zip(tbx.find_closest_facilities.__annotations__.keys(),
                        tbx.find_closest_facilities.__defaults__))
    if restrictions is None:
        restrictions = defaults['restrictions']
    if incidents is None:
        incidents = defaults['incidents']

    if facilities is None:
        facilities = defaults['facilities']

    if point_barriers is None:
        point_barriers = defaults['point_barriers']

    if line_barriers is None:
        line_barriers = defaults['line_barriers']

    if polygon_barriers is None:
        polygon_barriers = defaults['polygon_barriers']

    if attribute_parameter_values is None:
        attribute_parameter_values = defaults['attribute_parameter_values']

    if route_line_simplification_tolerance is None:
        route_line_simplification_tolerance = defaults['route_line_simplification_tolerance']

    from arcgis._impl.common._utils import inspect_function_inputs
    params = {
        "incidents" : incidents,
        "facilities" : facilities,
        "measurement_units" : measurement_units,
        "analysis_region" : analysis_region,
        "number_of_facilities_to_find" : number_of_facilities_to_find,
        "cutoff" : cutoff,
        "travel_direction" : travel_direction,
        "use_hierarchy" : use_hierarchy,
        "time_of_day" : time_of_day,
        "time_of_day_usage" : time_of_day_usage,
        "uturn_at_junctions" : uturn_at_junctions,
        "point_barriers" : point_barriers,
        "line_barriers" : line_barriers,
        "polygon_barriers" : polygon_barriers,
        "restrictions" : restrictions,
        "attribute_parameter_values" : attribute_parameter_values,
        "route_shape" : route_shape,
        "route_line_simplification_tolerance" : route_line_simplification_tolerance,
        "populate_directions" : populate_directions,
        "directions_language" : directions_language,
        "directions_distance_units" : directions_distance_units,
        "directions_style_name" : directions_style_name,
        "time_zone_for_time_of_day" : time_zone_for_time_of_day,
        "travel_mode" : travel_mode,
        "impedance" : impedance,
        "save_output_network_analysis_layer" : save_output_network_analysis_layer,
        "overrides" : overrides,
        "save_route_data" : save_route_data,
        "time_impedance" : time_impedance,
        "distance_impedance" : distance_impedance,
        "output_format" : output_format,
        "gis" : gis,
        "future" : True
    }
    params = inspect_function_inputs(tbx.find_closest_facilities, **params)
    job = tbx.find_closest_facilities(**params)
    if future:
        return job
    return job.result()

find_closest_facilities.__annotations__ = {
    'incidents': FeatureSet,
    'facilities': FeatureSet,
    'measurement_units': str,
    'analysis_region': str,
    'number_of_facilities_to_find': int,
    'cutoff': float,
    'travel_direction': str,
    'use_hierarchy': bool,
    'time_of_day': datetime,
    'time_of_day_usage': str,
    'uturn_at_junctions': str,
    'point_barriers': FeatureSet,
    'line_barriers': FeatureSet,
    'polygon_barriers': FeatureSet,
    'restrictions': str,
    'attribute_parameter_values': FeatureSet,
    'route_shape': str,
    'route_line_simplification_tolerance': LinearUnit,
    'populate_directions': bool,
    'directions_language': str,
    'directions_distance_units': str,
    'directions_style_name': str,
    'time_zone_for_time_of_day': str,
    'travel_mode': str,
    'impedance': str,
    'return': tuple}


import json as _json
from arcgis.raster._layer import ImageryLayer as _ImageryLayer
#from arcgis.raster._layer import Raster as _Raster
from arcgis.features import FeatureLayer as _FeatureLayer
import arcgis as _arcgis
import string as _string
import random as _random
from arcgis._impl.common._utils import _date_handler
import datetime
from arcgis.geometry import Geometry  as _Geometry
import numbers
import time
import os


import logging as _logging
_LOGGER = _logging.getLogger(__name__)

try:
    import numpy as _np
    import matplotlib.pyplot as _plt
    from matplotlib.pyplot import cm as _cm
except:
    pass


def _set_context(params, function_context = None):
    out_sr = _arcgis.env.out_spatial_reference
    process_sr = _arcgis.env.process_spatial_reference
    out_extent = _arcgis.env.analysis_extent
    mask = _arcgis.env.mask
    snap_raster = _arcgis.env.snap_raster
    cell_size = _arcgis.env.cell_size
    parallel_processing_factor = _arcgis.env.parallel_processing_factor

    context = {}

    if out_sr is not None:
        context['outSR'] = {'wkid': int(out_sr)}

    if out_extent is not None:
        context['extent'] = out_extent

    if process_sr is not None:
        context['processSR'] = {'wkid': int(process_sr)}


    if mask is not None:
        if isinstance(mask, _ImageryLayer):
            context['mask'] = {"url":mask._url}
        elif isinstance(mask,str):
            context['mask'] = {"url":mask}
    
    if cell_size is not None:
        if isinstance(cell_size, _ImageryLayer):
            context['cellSize'] = {"url":cell_size._url}
        elif isinstance(cell_size,str):
            if 'http:' in cell_size or 'https:' in cell_size:
                context['cellSize'] = {"url":cell_size}
            else:
                context['cellSize'] = cell_size
        else:
            context['cellSize'] = cell_size

    if snap_raster is not None:
        if isinstance(snap_raster, _ImageryLayer):
            context['snapRaster'] = {"url":snap_raster._url}
        elif isinstance(mask,str):
            context['snapRaster'] = {"url":snap_raster}


    if parallel_processing_factor is not None:
        context['parallelProcessingFactor'] = parallel_processing_factor


    if function_context is not None:
        if context is not None:
            context.update({k: function_context[k] for k in function_context.keys()})

        else:
            context = function_context

    if context:
        params["context"] = _json.dumps(context)

def _id_generator(size=6, chars=_string.ascii_uppercase + _string.digits):
    return ''.join(_random.choice(chars) for _ in range(size))

def _set_time_param(time):
    time_val = time
    if time is not None:
        if type(time) is list:
            if isinstance(time[0], datetime.datetime) or isinstance(time[0], datetime.date):
                if time[0].tzname() is None or time[0].tzname() != "UTC":
                    time[0] = time[0].astimezone(datetime.timezone.utc)
            if isinstance(time[1], datetime.datetime) or isinstance(time[1], datetime.date):
                if time[1].tzname() is None or time[1].tzname() != "UTC":
                    time[1] = time[1].astimezone(datetime.timezone.utc)
            starttime = _date_handler(time[0])
            endtime = _date_handler(time[1])
            if starttime is None:
                starttime = 'null'
            if endtime is None:
                endtime = 'null'
            time_val = "%s,%s" % (starttime, endtime)
        else:
            time_val = _date_handler(time)

    return time_val

def _to_datetime(dt):
    import datetime
    try:
        if dt<0:
            return datetime.datetime(1970, 1, 1) + datetime.timedelta(seconds=(dt/1000))
        else:
            return  datetime.datetime.utcfromtimestamp(dt/1000)
    except:
        return dt

def _datetime2ole(date):
    #date = datetime.strptime(date, '%d-%b-%Y')
    import datetime
    OLE_TIME_ZERO = datetime.datetime(1899, 12, 30)
    delta = date - OLE_TIME_ZERO
    return float(delta.days) + (float(delta.seconds) / 86400)

def _ole2datetime(oledt):
    import datetime
    OLE_TIME_ZERO = datetime.datetime(1899, 12, 30, 0, 0, 0)
    try:
        return OLE_TIME_ZERO + datetime.timedelta(days=float(oledt))
    except:
        return datetime.datetime.utcfromtimestamp(oledt/1000)

def _iso_to_datetime(timestamp):
    format_string = '%Y-%m-%dT%H:%M:%S%z'
    try:
        colon = timestamp[-3]
        colonless_timestamp = timestamp
        if colon == ':':
            colonless_timestamp = timestamp[:-3] + timestamp[-2:]
        dt_ob = datetime.datetime.strptime(colonless_timestamp, format_string)
        return dt_ob.replace(tzinfo=None)
    except:
        try:
            format_string = '%Y-%m-%dT%H:%M:%S'
            dt_ob = datetime.datetime.strptime(timestamp, format_string)
            return dt_ob
        except:
            return timestamp

def _check_if_iso_format(timestamp):
    format_string = '%Y-%m-%dT%H:%M:%S%z'
    try:
        colon = timestamp[-3]
        colonless_timestamp = timestamp
        if colon == ':':
            colonless_timestamp = timestamp[:-3] + timestamp[-2:]
        dt_ob = datetime.datetime.strptime(colonless_timestamp, format_string)
        return True
    except:
        try:
            format_string = '%Y-%m-%dT%H:%M:%S'
            dt_ob = datetime.datetime.strptime(timestamp, format_string)
            return dt_ob
        except:
            return False

def _time_filter(time_extent,ele):
    if time_extent is not None:
        if isinstance(time_extent, datetime.datetime):
            if(ele<time_extent):
                return True
            else:
                return False
        elif isinstance(time_extent, list):
            if isinstance(time_extent[0], datetime.datetime) and isinstance(time_extent[1], datetime.datetime):                                                
                if(time_extent[0] < ele and ele < time_extent[1]):
                    return True
                else:
                    return False

        else:                            
            return True
    else:
        return True


def _linear_regression(sample_size, date_list, x, y):
    ncoefficient = 2
    if sample_size < ncoefficient:
        _LOGGER.warning("Trend line cannot be drawn. Insufficient points to plot Linear Trend Line")
        return [],[]

    AA = _np.empty([sample_size,ncoefficient], dtype=float, order='C')
    BB = _np.empty([sample_size,1], dtype=float, order='C')
    XX = _np.empty([ncoefficient,1], dtype=float, order='C')
    for i in range(sample_size):
        n=0
        AA[i][n] = date_list[i] 
        AA[i][n+1] = 1
        BB[i] = y[i]

    x1 = _np.linalg.lstsq(AA, BB, rcond=None)[0]

    YY=[]
    for i in range(sample_size):
        y_temp=x1[0][0]*date_list[i] + x1[1][0]
        YY.append(y_temp)
    return x,YY


def _harmonic_regression(sample_size, date_list, x, y, trend_order):
    PI2_Year = 3.14159265*2/365.25

    ncoefficient = 2 * (trend_order + 1)
    if sample_size < ncoefficient:
        _LOGGER.warning("Trend line cannot be drawn. Insufficient points to plot Harmonic Trend Line for trend order "+str(trend_order)+". Please try specifying a lower trend order.")
        return [],[]

    AA = _np.empty([sample_size,ncoefficient], dtype=float, order='C')
    BB = _np.empty([sample_size,1], dtype=float, order='C')
    XX = _np.empty([ncoefficient,1], dtype=float, order='C')

    for i in range(sample_size):
        n=0
        AA[i][n] = date_list[i] 
        AA[i][n+1] = 1

        for j in range(1,trend_order+1):
            AA[i][n + 2 * j] = _np.sin(PI2_Year * j * date_list[i])
            AA[i][n + 2 * j + 1] = _np.cos(PI2_Year * j * date_list[i])

        BB[i] = y[i]

    x1 = _np.linalg.lstsq(AA, BB, rcond=None)[0]
    YY=[]
    for i in range(sample_size):
        y_temp=x1[0][0]*date_list[i] + x1[1][0]
        for q in range(2,len(x1),2):
            y_temp=y_temp + x1[q][0] * _np.sin(2 * 3.14159265358979323846 * (q / 2) * date_list[i] / 365.25)
            y_temp=y_temp + x1[q+1][0] * _np.cos(2 * 3.14159265358979323846 * (q / 2) * date_list[i] / 365.25)
        YY.append(y_temp)
    return x, YY

def _epoch_to_iso(dt):
    import datetime
    try:
        if dt<0:
            return (datetime.datetime(1970, 1, 1, tzinfo = datetime.timezone.utc) + datetime.timedelta(seconds=(dt/1000))).isoformat()
        else:
            return  datetime.datetime.fromtimestamp(dt/1000, tz=datetime.timezone.utc).isoformat()
    except:
        return dt

def _datetime2ole(date):
    #date = datetime.strptime(date, '%d-%b-%Y')
    import datetime
    OLE_TIME_ZERO = datetime.datetime(1899, 12, 30)
    delta = date - OLE_TIME_ZERO
    return float(delta.days) + (float(delta.seconds) / 86400)

def _ole2datetime(oledt):
    import datetime
    OLE_TIME_ZERO = datetime.datetime(1899, 12, 30, 0, 0, 0)
    try:
        return OLE_TIME_ZERO + datetime.timedelta(days=float(oledt))
    except:
        return datetime.datetime.utcfromtimestamp(oledt/1000)

def _iso_to_datetime(timestamp):
    format_string = '%Y-%m-%dT%H:%M:%S%z'
    try:
        colon = timestamp[-3]
        colonless_timestamp = timestamp
        if colon == ':':
            colonless_timestamp = timestamp[:-3] + timestamp[-2:]
        dt_ob = datetime.datetime.strptime(colonless_timestamp, format_string)
        return dt_ob.replace(tzinfo=None)
    except:
        try:
            format_string = '%Y-%m-%dT%H:%M:%S.%f%z'
            dt_ob = datetime.datetime.strptime(colonless_timestamp, format_string)
            return dt_ob.replace(tzinfo=None)
        except:
            try:
                format_string = '%Y-%m-%dT%H:%M:%S'
                dt_ob = datetime.datetime.strptime(timestamp, format_string)
                return dt_ob
            except:
                return timestamp

def _check_if_iso_format(timestamp):
    format_string = '%Y-%m-%dT%H:%M:%S%z'
    try:
        colon = timestamp[-3]
        colonless_timestamp = timestamp
        if colon == ':':
            colonless_timestamp = timestamp[:-3] + timestamp[-2:]
        dt_ob = datetime.datetime.strptime(colonless_timestamp, format_string)
        return True
    except:
        try:
            format_string = '%Y-%m-%dT%H:%M:%S'
            dt_ob = datetime.datetime.strptime(timestamp, format_string)
            return dt_ob
        except:
            return False

def _local_function_template(operation_number=None):
    template_dict = {
  "name" : "max_rft",
  "description" : "A raster function template.",
  "function" : {
    "pixelType" : "UNKNOWN",
    "name" : "Cell Statistics",
    "description" : "Calculates a per-cell statistic from multiple rasters.  The available statistics are Majority, Maximum, Mean, Median, Minimum, Minority, Range, Standard Deviation, Sum, and Variety.",
    "type" : "LocalFunction",
    "_object_id" : 1
  },
  "arguments" : {
    "Rasters" : {
      "name" : "Rasters",
      "value" : {
        "elements" : [
        ],
        "type" : "ArgumentArray",
        "_object_id" : 2
      },
      "aliases" : [
        "__IsRasterArray__"
      ],
      "isDataset" : False,
      "isPublic" : False,
      "type" : "RasterFunctionVariable",
      "_object_id" : 3
    },
    "Operation" : {
      "name" : "Operation",
      "value" : "",
      "isDataset" : False,
      "isPublic" : False,
      "type" : "RasterFunctionVariable",
      "_object_id" : 4
    },
    "CellsizeType" : {
      "name" : "CellsizeType",
      "value" : 2,
      "isDataset" : False,
      "isPublic" : False,
      "type" : "RasterFunctionVariable",
      "_object_id" : 5
    },
    "ExtentType" : {
      "name" : "ExtentType",
      "value" : 1,
      "isDataset" : False,
      "isPublic" : False,
      "type" : "RasterFunctionVariable",
      "_object_id" : 6
    },
    "ProcessAsMultiband" : {
      "name" : "ProcessAsMultiband",
      "value" : True,
      "isDataset" : False,
      "isPublic" : False,
      "type" : "RasterFunctionVariable",
      "_object_id" : 7
    },
    "MatchVariable" : {
      "name" : "MatchVariable",
      "value" : True,
      "isDataset" : False,
      "isPublic" : False,
      "type" : "RasterFunctionVariable",
      "_object_id" : 8
    },
    "UnionDimension" : {
      "name" : "UnionDimension",
      "value" : False,
      "isDataset" : False,
      "isPublic" : False,
      "type" : "RasterFunctionVariable",
      "_object_id" : 9
    },
    "type" : "LocalFunctionArguments",
    "_object_id" : 10
  },
  "functionType" : 0,
  "thumbnail" : ""
}
    if operation_number is not None:
        template_dict["arguments"]["Operation"]["value"]=operation_number
    return template_dict

def _get_geometry(data):
    if data is None:
        return None

    if isinstance(data, _Geometry):
        return data
    elif isinstance(data, _arcgis.raster.Raster):
        return _Geometry(data.extent)
    elif isinstance(data, _ImageryLayer):
        return _Geometry(data.extent)
    elif isinstance(data, _FeatureLayer):
        return _get_geometry_from_feature_layer(data)
    else:
        return data

def _get_geometry_from_feature_layer(data):
    geo=None
    layer_fset = layer.query()
    for ele in layer_fset.features:
        geo = geo.union(_Geometry(ele.geometry)) if geo else _Geometry(ele.geometry)
    return geometry

def build_query_string(field_name, operator, field_values):
    operator_map = {
    "equals": "=",
    "less_than": "<",
    "greater_than": ">",
    "not_equals": "<>",
    "not_less_than": ">=",
    "not_greater_than": "<=",
        }

    if operator in operator_map:
        if isinstance(field_values, numbers.Number):
            return field_name + ' ' + operator_map[operator] + ' ' + str(field_values)
        elif isinstance(field_values, str):
            return field_name + ' ' + operator_map[operator] + ' \'' + field_values + '\''
        else:
            raise TypeError('field_value must be numeric or string')

    elif operator in ['starts_with', 'ends_with', 'not_starts_with', 'not_ends_with', 'contains', 'not_contains']:
        if not isinstance(field_values, str):
            raise TypeError('field_value must be string')
        if operator == 'starts_with':
            return field_name + ' LIKE ' + '\'' + field_values + '%\''
        elif operator == 'ends_with':
            return field_name + ' LIKE' + '\'%' + field_values + '\''
        elif operator == 'not_starts_with':
            return field_name + ' NOT LIKE ' + '\'' + field_values + '%\''
        elif operator == 'not_ends_with':
            return field_name + ' NOT LIKE ' + '\'%' + field_values + '\''
        elif operator == 'contains':
            return field_name + ' LIKE ' + '\'%' + field_values + '%\''
        elif operator == 'not_contains':
            return field_name + ' NOT LIKE ' + '\'%' + field_values + '%\''
    elif operator == 'in':
        if not isinstance(field_values, list):
            raise TypeError('field_values must be type list for operator "in"')
        values = '('
        for item in field_values:
            if not (isinstance(item, numbers.Number) or isinstance(item, str)):
                raise TypeError('item in field_values must be numeric or string')
            if values == '(':
                values += '\'' + item + '\'' if isinstance(item, str) else str(item)
            else:
                values += ',\'' + item + '\'' if isinstance(item, str) else ',' + str(item)
        values += ')'
        return field_name + ' IN ' + values
    elif operator == 'not_in':
        values = '('
        for item in field_values:
            if not (isinstance(item, numbers.Number) or isinstance(item, str)):
                raise TypeError('item in field_values must be numeric or string')
            if values == '(':
                values += '\'' + item + '\'' if isinstance(item, str) else str(item)
            else:
                values += ',\'' + item + '\'' if isinstance(item, str) else ',' + str(item)
        values += ')'
        return field_name + ' NOT IN ' + values
    else:
        raise ValueError('invalid operator value')

def _generate_direct_access_url(gis=None):
    """helper fn to get the direct access url for azure storage"""
    gis = _arcgis.env.active_gis if gis is None else gis
    url = "%s/sharing/rest/content/users/%s/generateDirectAccessUrl" % (gis._portal.url,
                                                                 gis._username)
    params = {"f" : "json", "storeType":"rasterStore"}
    res = gis._portal.con.post(url, params)
    if isinstance(res, dict):
        if "url" in res.keys():
            return res["url"]
        else:
            raise RuntimeError("Couldn't generate direct access url")
    else:
        raise RuntimeError("Couldn't generate direct access url")
    
def _upload_imagery_agol(files, gis=None):
    """uploads a file to the image layer to AGOL and returns the list of urls"""

    try:
        from azure.storage.blob import ContainerClient
    except:
        print("Install Azure library packages for Python. (version - azure-storage-blob-12.5.0) \
        (https://docs.microsoft.com/en-us/azure/developer/python/azure-sdk-install)")
    sas_url = _generate_direct_access_url(gis)
    container = ContainerClient.from_container_url(sas_url)
    if not isinstance(files,list):
        files = [files]

    url_list = []
    for file in files:
        current_time = int(time.time())
        prefix =  "_images/"+str(current_time)+"/"

        if os.path.exists(file):
            if(os.path.isdir(file)):
                folder = os.path.basename(file)
                basename_len=len(os.path.dirname(file))
                for root,d_names,f_names in os.walk(file):
                    for f in f_names:
                        blobname = prefix + (root+"/"+f)[basename_len+1:].replace(os.sep, '/')
                        filepath = os.path.join(root, f)
                        blob=container.get_blob_client(blobname)
                        url = blob.url.split("?", 1)[0]
                        url_list.append(url)
                        with open(filepath, "rb") as data:
                            blob.upload_blob(data, blob_type="BlockBlob")

            else:
                blobname = prefix+os.path.basename(file).replace(os.sep, '/')
                blob=container.get_blob_client(blobname)
                with open(file, "rb") as data:
                    blob.upload_blob(data, blob_type="BlockBlob")
                url = blob.url.split("?", 1)[0]
                url_list.append(url)

    return url_list

def _upload_imagery_enterprise(files, raster_type_name=None, gis=None):
    """uploads a file to the image layer to enterprise and returns the item id"""
    
    ra_url = gis.properties.helperServices["rasterAnalytics"]["url"]
    url = "%s/uploads/upload" % ra_url
    params = {
        "f" : 'json'
    }

   
    if not isinstance(files,list):
        files = [files]

    item_ids_list = []
    res = {}
    
    append_path = False
    for file in files:
        item_id_dict={}
        if os.path.exists(file):
            if(os.path.isdir(file)):
                if file.endswith(".crf") or raster_type_name !="Raster Dataset":
                    append_path = True
                folder = os.path.basename(file)
                basename_len=len(os.path.dirname(file))
                for root,d_names,f_names in os.walk(file):
                    for f in f_names:
                        fp =os.path.join(root, f)
                        path = ("/"+root+"/"+f)[basename_len+1:].replace(os.sep, '/')
                        files_param = {'file' : fp }
                        try:
                            res = gis._con.post(path=url, postdata=params, files=files_param)
                        except Exception as e:
                            _LOGGER.warning('file: '+str(fp)+ " "+ str(e))

                        if 'success' in res and res['success']:
                            item_id = res['item']['itemID']
                            res = {}
                            if append_path:
                                item_id_dict = {"itemId":item_id, "path":fp}
                                item_ids_list.append(item_id_dict)
                                item_id_dict={}
                            else:
                                item_ids_list.append(item_id)

            else:
                files_param = {'file' : file}
                try:
                    res = gis._con.post(path=url, postdata=params, files=files_param)
                except Exception as e:
                    _LOGGER.warning('file: '+str(file)+ " "+ str(e))
                if 'success' in res and res['success']:
                    item_id = res['item']['itemID']
                    res = {}
                if item_id is not None:
                    item_ids_list.append(item_id)

    return item_ids_list



import torch
import torch.nn as nn
import torch.nn.functional as F
from fastai.torch_core import Module
from functools import partial
from . import *


device = 'cuda' if torch.cuda.is_available() else 'cpu'


def noop(x): return x


class LambdaPlus(Module):
    def __init__(self, func, *args, **kwargs): self.func, self.args, self.kwargs = func, args, kwargs

    def forward(self, x): return self.func(x, *self.args, **self.kwargs)


def get_act_layer(act_fn, **kwargs):
    act_fn = act_fn.lower()
    assert act_fn in ['relu', 'leakyrelu', 'prelu', 'elu'], 'incorrect act_fn'
    if act_fn == 'relu':
        return nn.ReLU()
    elif act_fn == 'leakyrelu':
        return nn.LeakyReLU(**kwargs)
    elif act_fn == 'prelu':
        return nn.PReLU(**kwargs)
    elif act_fn == 'elu':
        return nn.ELU(**kwargs)


def same_padding1d(seq_len, ks, stride=1, dilation=1):
    assert stride > 0
    assert dilation >= 1
    effective_ks = (ks - 1) * dilation + 1
    out_dim = (seq_len + stride - 1) // stride
    p = max(0, (out_dim - 1) * stride + effective_ks - seq_len)
    padding_before = p // 2
    padding_after = p - padding_before
    return padding_before, padding_after


class ZeroPad1d(nn.ConstantPad1d):
    def __init__(self, padding):
        super().__init__(padding, 0.)


class ConvSP1d(nn.Module):
    "Conv1d padding='same'"

    def __init__(self, c_in, c_out, ks, stride=1, padding='same', dilation=1, bias=True):
        super().__init__()
        self.ks, self.stride, self.dilation = ks, stride, dilation
        self.conv = nn.Conv1d(c_in, c_out, ks, stride=stride, padding=0, dilation=dilation, bias=bias)
        self.zeropad = ZeroPad1d
        self.weight = self.conv.weight
        self.bias = self.conv.bias

    def forward(self, x):
        padding = same_padding1d(x.shape[-1], self.ks, stride=self.stride, dilation=self.dilation)
        return self.conv(self.zeropad(padding)(x))


def convlayer(c_in, c_out, ks=3, padding='same', bias=True, stride=1,
              bn_init=False, zero_bn=False, bn_before=True,
              act_fn='relu', **kwargs):
    '''conv layer (padding="same") + bn + act'''
    if ks % 2 == 1 and padding == 'same': padding = ks // 2
    layers = [ConvSP1d(c_in, c_out, ks, bias=bias, stride=stride) if padding == 'same' else \
                  nn.Conv1d(c_in, c_out, ks, stride=stride, padding=padding, bias=bias)]
    bn = nn.BatchNorm1d(c_out)
    if bn_init: nn.init.constant_(bn.weight, 0. if zero_bn else 1.)
    if bn_before: layers.append(bn)
    if act_fn: layers.append(get_act_layer(act_fn, **kwargs))
    if not bn_before: layers.append(bn)
    return nn.Sequential(*layers)


class Flatten(nn.Module):
    def forward(self, x): return x.view(x.size(0), -1)


class Squeeze(nn.Module):
    def __init__(self, dim=-1):
        super().__init__()
        self.dim = dim

    def forward(self, x): return x.squeeze(dim=self.dim)


class Unsqueeze(nn.Module):
    def __init__(self, dim=-1):
        super().__init__()
        self.dim = dim

    def forward(self, x): return x.unsqueeze(dim=self.dim)


class YRange(nn.Module):
    def __init__(self, y_range: tuple):
        super().__init__()
        self.y_range = y_range

    def forward(self, x):
        x = F.sigmoid(x)
        x = x * (self.y_range[1] - self.y_range[0])
        return x + self.y_range[0]


class Mult(nn.Module):
    def __init__(self, mult, trainable=True):
        self.mult, self.trainable = mult, trainable
        super().__init__()
        self.weight = nn.Parameter(torch.Tensor(1).fill_(self.mult).to(device), requires_grad=self.trainable)

    def forward(self, x):
        return x.mul_(self.weight)


class Exp(nn.Module):
    def __init__(self, exp, trainable=True):
        self.exp, self.trainable = exp, trainable
        super().__init__()
        self.weight = nn.Parameter(torch.Tensor(1).fill_(self.exp).to(device), requires_grad=self.trainable)

    def forward(self, x):
        return x ** self.weight


class GeneralRelu(nn.Module):
    def __init__(self, leak=None, sub=0., maxv=None, **kwargs):
        super().__init__()
        self.leak, self.sub, self.maxv = leak, sub, maxv

    def forward(self, x):
        x = F.leaky_relu(x, self.leak) if self.leak is not None else F.relu(x)
        x = x.sub_(self.sub)
        if self.maxv is not None: x.clamp_max_(self.maxv)
        return x


def get_act_fn_norm(act_fn):
    x = torch.randn(1000000)
    x = act_fn(x)
    x1 = x / x.std()
    x2 = x1 - x1.mean()
    return 1 / x.std(), x1.mean()


class AFN(nn.Module):
    def __init__(self, act_fn=F.relu, trainable=True, **kwargs):
        super().__init__()
        self.act_fn = partial(act_fn, **kwargs)
        mul, add = get_act_fn_norm(self.act_fn)
        self.weight = nn.Parameter(torch.Tensor(1).fill_(mul).to(device), requires_grad=trainable)
        self.bias = nn.Parameter(torch.Tensor(1).fill_(add).to(device), requires_grad=trainable)

    def forward(self, x):
        x = self.act_fn(x)
        x.mul_(self.weight)
        x.add_(self.bias)
        return x


from ._arcgis_model import ArcGISModel
from pathlib import Path
import json
from ._arcgis_model import _EmptyData, _change_tail
from ._codetemplate import instance_detector_prf
import math

try:
    import torch
    from fastai.vision.learner import cnn_learner
    from fastai.callbacks.hooks import model_sizes
    from fastai.vision.learner import create_body
    from fastai.vision.image import open_image
    from fastai.vision import flatten_model
    from fastai.core import has_arg, split_kwargs_by_func
    from torchvision.models import resnet34
    from torchvision import models
    import numpy as np
    from .._data import prepare_data, _raise_fastai_import_error
    from fastai.callbacks import EarlyStoppingCallback
    from ._arcgis_model import SaveModelCallback, _set_multigpu_callback, _get_backbone_meta, _resnet_family, _set_ddp_multigpu, _isnotebook
    import torchvision
    from torchvision import models
    from torchvision.models.detection.faster_rcnn import FastRCNNPredictor
    from torchvision.models.detection.mask_rcnn import MaskRCNNPredictor
    from fastai.basic_train import Learner
    from ._maskrcnn_utils import is_no_color, mask_rcnn_loss, train_callback, compute_class_AP
    from ._MaskRCNN_PointRend import create_pointrend
    from .._utils.common import get_multispectral_data_params_from_emd, _get_emd_path
    from fastai.torch_core import split_model_idx
    import matplotlib.pyplot as plt
    import matplotlib.patches as patches
    import matplotlib
    from fastai.basic_data import DatasetType
    from torchvision.models.detection.backbone_utils import resnet_fpn_backbone
    import os as arcgis_os
    from .._utils.common import get_nbatches, image_batch_stretcher

    HAS_FASTAI = True
except Exception as e:
    #raise Exception(e)
    HAS_FASTAI = False


class MaskRCNN(ArcGISModel):
    """
    Model architecture from https://arxiv.org/abs/1703.06870.
    Creates a ``MaskRCNN`` Instance segmentation model,
    based on https://github.com/pytorch/vision/blob/master/torchvision/models/detection/mask_rcnn.py.

    =====================   ===========================================
    **Argument**            **Description**
    ---------------------   -------------------------------------------
    data                    Required fastai Databunch. Returned data object from
                            ``prepare_data`` function.
    ---------------------   -------------------------------------------
    backbone                Optional function. Backbone CNN model to be used for
                            creating the base of the `MaskRCNN`, which
                            is `resnet50` by default. 
                            Compatible backbones: 'resnet18', 'resnet34', 'resnet50', 'resnet101', 'resnet152'
    ---------------------   -------------------------------------------
    pretrained_path         Optional string. Path where pre-trained model is
                            saved.
    ---------------------   -------------------------------------------
    pointrend               Optional boolean. If True, it will use PointRend
                            architecture on top of the segmentation head.
                            Default: False. PointRend architecture from
                            https://arxiv.org/pdf/1912.08193.pdf.      
    =====================   ===========================================
    
    **kwargs**

    =============================   =============================================
    **Argument**                    **Description**
    -----------------------------   ---------------------------------------------
    rpn_pre_nms_top_n_train         Optional int. Number of proposals to keep before
                                    applying NMS during training.
                                    Default: 2000
    -----------------------------   ---------------------------------------------
    rpn_pre_nms_top_n_test          Optional int. Number of proposals to keep before
                                    applying NMS during testing.
                                    Default: 1000
    -----------------------------   ---------------------------------------------
    rpn_post_nms_top_n_train        Optional int. Number of proposals to keep after
                                    applying NMS during training.
                                    Default: 2000
    -----------------------------   ---------------------------------------------
    rpn_post_nms_top_n_test         Optional int. Number of proposals to keep after
                                    applying NMS during testing.
                                    Default: 1000
    -----------------------------   ---------------------------------------------
    rpn_nms_thresh                  Optional float. NMS threshold used for postprocessing
                                    the RPN proposals.
                                    Default: 0.7
    -----------------------------   ---------------------------------------------
    rpn_fg_iou_thresh               Optional float. Minimum IoU between the anchor
                                    and the GT box so that they can be considered
                                    as positive during training of the RPN.
                                    Default: 0.7
    -----------------------------   ---------------------------------------------
    rpn_bg_iou_thresh               Optional float. Maximum IoU between the anchor and
                                    the GT box so that they can be considered as negative
                                    during training of the RPN.
                                    Default: 0.3
    -----------------------------   ---------------------------------------------
    rpn_batch_size_per_image        Optional int. Number of anchors that are sampled
                                    during training of the RPN for computing the loss.
                                    Default: 256
    -----------------------------   ---------------------------------------------
    rpn_positive_fraction           Optional float. Proportion of positive anchors in a
                                    mini-batch during training of the RPN.
                                    Default: 0.5
    -----------------------------   ---------------------------------------------
    box_score_thresh                Optional float. During inference, only return proposals
                                    with a classification score greater than box_score_thresh
                                    Default: 0.05
    -----------------------------   ---------------------------------------------
    box_nms_thresh                  Optional float. NMS threshold for the prediction head.
                                    Used during inference.
                                    Default: 0.5
    -----------------------------   ---------------------------------------------
    box_detections_per_img          Optional int. Maximum number of detections per
                                    image, for all classes.
                                    Default: 100
    -----------------------------   ---------------------------------------------
    box_fg_iou_thresh               Optional float. Minimum IoU between the proposals and
                                    the GT box so that they can be considered as positive
                                    during training of the classification head.
                                    Default: 0.5
    -----------------------------   ---------------------------------------------
    box_bg_iou_thresh               Optional float. Maximum IoU between the proposals and 
                                    the GT box so that they can be considered as negative 
                                    during training of the classification head.
                                    Default: 0.5
    -----------------------------   ---------------------------------------------
    box_batch_size_per_image        Optional int. Number of proposals that are sampled during
                                    training of the classification head.
                                    Default: 512
    -----------------------------   ---------------------------------------------
    box_positive_fraction           Optional float. Proportion of positive proposals in a
                                    mini-batch during training of the classification head.
                                    Default: 0.25
    =============================   =============================================

    :returns: ``MaskRCNN`` Object
    """
    def __init__(self, data, backbone=None, pretrained_path=None, pointrend=False, *args, **kwargs):

        # Set default backbone to be 'resnet50'
        if backbone is None:
            backbone = models.resnet50

        self._check_dataset_support(data)
        if not (self._check_backbone_support(backbone)):
            raise Exception (f"Enter only compatible backbones from {', '.join(self.supported_backbones)}")

        super().__init__(data, backbone, **kwargs)
        if self._is_multispectral:
            self._backbone_ms = self._backbone
            self._backbone = self._orig_backbone
            scaled_mean_values = data._scaled_mean_values[data._extract_bands].tolist()
            scaled_std_values = data._scaled_std_values[data._extract_bands].tolist()

        self._code = instance_detector_prf

        self._pointrend = pointrend

        self.maskrcnn_kwargs, kwargs = split_kwargs_by_func(kwargs, models.detection.MaskRCNN.__init__)

        if self._backbone.__name__ is 'resnet50':
            model = models.detection.maskrcnn_resnet50_fpn(
                pretrained=True,
                min_size = 1.5*data.chip_size,
                max_size = 2*data.chip_size,
                **self.maskrcnn_kwargs
            )

            if self._is_multispectral:
                model.backbone = _change_tail(model.backbone, data)
                model.transform.image_mean = scaled_mean_values
                model.transform.image_std = scaled_std_values
        elif self._backbone.__name__ in ['resnet18','resnet34']:
            if self._is_multispectral:
                backbone_small = create_body(self._backbone_ms, cut=_get_backbone_meta(self._backbone.__name__)['cut'])
                backbone_small.out_channels = 512
                model = models.detection.MaskRCNN(
                    backbone_small, 
                    91, 
                    min_size = 1.5*data.chip_size, 
                    max_size = 2*data.chip_size, 
                    image_mean = scaled_mean_values, 
                    image_std = scaled_std_values,
                    **self.maskrcnn_kwargs
                )
            else:
                backbone_small = create_body(self._backbone)
                backbone_small.out_channels = 512
                model = models.detection.MaskRCNN(
                    backbone_small,
                    91,
                    min_size = 1.5*data.chip_size,
                    max_size = 2*data.chip_size,
                    **self.maskrcnn_kwargs
                )
        else:
            backbone_fpn = resnet_fpn_backbone(self._backbone.__name__, True)
            if self._is_multispectral:
                backbone_fpn = _change_tail(backbone_fpn, data)
                model = models.detection.MaskRCNN(
                    backbone_fpn, 
                    91, 
                    min_size = 1.5*data.chip_size, 
                    max_size = 2*data.chip_size, 
                    image_mean = scaled_mean_values, 
                    image_std = scaled_std_values,
                    **self.maskrcnn_kwargs
                )
            else:
                model = models.detection.MaskRCNN(
                    backbone_fpn,
                    91,
                    min_size = 1.5*data.chip_size,
                    max_size = 2*data.chip_size,
                    **self.maskrcnn_kwargs
                )

        in_features = model.roi_heads.box_predictor.cls_score.in_features
        model.roi_heads.box_predictor = FastRCNNPredictor(in_features, data.c)

        if pointrend:
            model = create_pointrend(model, data.c)
        else:
            in_features_mask = model.roi_heads.mask_predictor.conv5_mask.in_channels
            hidden_layer = 256
            model.roi_heads.mask_predictor = MaskRCNNPredictor(in_features_mask,
                                                        hidden_layer,
                                                        data.c)

        if not _isnotebook() and arcgis_os.name=='posix':
            _set_ddp_multigpu(self)
            if self._multigpu_training:
                self.learn = Learner(data, model, loss_func=mask_rcnn_loss).to_distributed(self._rank_distributed)
            else:
                self.learn = Learner(data, model, loss_func=mask_rcnn_loss)
        else:
            self.learn = Learner(data, model, loss_func=mask_rcnn_loss)
        self.learn.callbacks.append(train_callback(self.learn))
        self.learn.model = self.learn.model.to(self._device)
        self.learn.c_device = self._device

        # fixes for zero division error when slice is passed
        idx = 27
        if self._backbone.__name__ in ['resnet18','resnet34']:
            idx = self._freeze()
        self.learn.layer_groups = split_model_idx(self.learn.model, [idx])
        self.learn.create_opt(lr=3e-3)

        # make first conv weights learnable
        self._arcgis_init_callback()

        if pretrained_path is not None:
            self.load(pretrained_path)
                
        if self._is_multispectral:
            self._orig_backbone = self._backbone
            self._backbone = self._backbone_ms
             
    def unfreeze(self):
        for _, param in self.learn.model.named_parameters():
            param.requires_grad = True

    def _freeze(self):
        "Freezes the pretrained backbone."
        for idx, i in enumerate(flatten_model(self.learn.model.backbone)):
            if isinstance(i, (torch.nn.BatchNorm2d)):
                continue
            for p in i.parameters():
                p.requires_grad = False
        return idx

    def __str__(self):
        return self.__repr__()

    def __repr__(self):
        return '<%s>' % (type(self).__name__)

    @property
    def supported_backbones(self):
        """ Supported torchvision backbones for this model. """        
        return MaskRCNN._supported_backbones()

    @staticmethod
    def _supported_backbones():
        return [*_resnet_family]

    @property
    def  supported_datasets(self):
        """ Supported dataset types for this model. """
        return MaskRCNN._supported_datasets()
    
    @staticmethod
    def _supported_datasets():
        return ['RCNN_Masks'] 
    
    @classmethod
    def from_model(cls, emd_path, data=None, **kwargs):
        """
        Creates a ``MaskRCNN`` Instance segmentation object from an Esri Model Definition (EMD) file.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        emd_path                Required string. Path to Deep Learning Package
                                (DLPK) or Esri Model Definition(EMD) file.
        ---------------------   -------------------------------------------
        data                    Required fastai Databunch or None. Returned data
                                object from ``prepare_data`` function or None for
                                inferencing.

        =====================   ===========================================

        :returns: `MaskRCNN` Object
        """

        emd_path = _get_emd_path(emd_path)
        with open(emd_path) as f:
            emd = json.load(f)
            
        model_file = Path(emd['ModelFile'])
        
        if not model_file.is_absolute():
            model_file = emd_path.parent / model_file
            
        model_params = emd['ModelParameters']
        maskrcnn_kwargs = emd.get("MaskRCNNkwargs", {})

        try:
            class_mapping = {i['Value'] : i['Name'] for i in emd['Classes']}
            color_mapping = {i['Value'] : i['Color'] for i in emd['Classes']}
        except KeyError:
            class_mapping = {i['ClassValue'] : i['ClassName'] for i in emd['Classes']} 
            color_mapping = {i['ClassValue'] : i['Color'] for i in emd['Classes']}                

        if data is None:
            data = _EmptyData(path=emd_path.parent.parent, loss_func=None, c=len(class_mapping) + 1, chip_size=kwargs.get('chip_size', emd['ImageHeight']))
            data.class_mapping = class_mapping
            data.color_mapping = color_mapping
            data.emd_path = emd_path
            data.emd = emd
            data = get_multispectral_data_params_from_emd(data, emd)

        return cls(data, **model_params, pretrained_path=str(model_file), **maskrcnn_kwargs)

    def _get_emd_params(self, save_inference_file):
        import random

        _emd_template = {"ModelParameters" : {}}
        _emd_template["Framework"] = "arcgis.learn.models._inferencing"
        _emd_template["ModelConfiguration"] = "_maskrcnn_inferencing"
        if save_inference_file:
            _emd_template["InferenceFunction"] = "ArcGISInstanceDetector.py"
        else:
            _emd_template["InferenceFunction"] = "[Functions]System\\DeepLearning\\ArcGISLearn\\ArcGISInstanceDetector.py"
        _emd_template["ModelType"] = "InstanceDetection"
        _emd_template["MaskRCNNkwargs"] = self.maskrcnn_kwargs
        _emd_template["ModelParameters"]["pointrend"] = self._pointrend
        _emd_template["ExtractBands"] = [0, 1, 2]
        _emd_template["SupportsVariableTileSize"] = True
        _emd_template['Classes'] = []
        class_data = {}
        for i, class_name in enumerate(self._data.classes[1:]):  # 0th index is background
            inverse_class_mapping = {v: k for k, v in self._data.class_mapping.items()}
            class_data["Value"] = inverse_class_mapping[class_name]
            class_data["Name"] = class_name
            color = [random.choice(range(256)) for i in range(3)] if is_no_color(self._data.color_mapping) else \
            self._data.color_mapping[inverse_class_mapping[class_name]]
            class_data["Color"] = color
            _emd_template['Classes'].append(class_data.copy())

        return _emd_template

    @property
    def _model_metrics(self):
        return {'average_precision_score': self.average_precision_score(show_progress=True)}

    def _predict_results(self, xb):

        self.learn.model.eval()
        xb_l = xb.to(self._device)
        predictions = self.learn.model(list(xb_l))
        xb_l = xb_l.detach().cpu()
        del xb_l
        predictionsf = []
        for i in range(len(predictions)):
            predictionsf.append({})
            predictionsf[i]['masks'] = predictions[i]['masks'].detach().cpu().numpy()
            predictionsf[i]['boxes'] = predictions[i]['boxes'].detach().cpu().numpy()
            predictionsf[i]['labels'] = predictions[i]['labels'].detach().cpu().numpy()
            predictionsf[i]['scores'] = predictions[i]['scores'].detach().cpu().numpy()
            del predictions[i]['masks']
            del predictions[i]['boxes']
            del predictions[i]['labels']
            del predictions[i]['scores']
        if self._device == torch.device('cuda'):
            torch.cuda.empty_cache()
        return predictionsf

    def _predict_postprocess(self, predictions, threshold=0.5, box_threshold = 0.5):

        pred_mask = []
        pred_box = []

        for i in range(len(predictions)):
            out = predictions[i]['masks'].squeeze()
            pred_box.append([])

            if out.shape[0] != 0:  # handle for prediction with n masks
                if len(out.shape) == 2: # for out dimension hxw (in case of only one predicted mask)
                    out = out[None]
                ymask = np.where(out[0]> threshold, 1, 0)
                if predictions[i]['scores'][0] > box_threshold:
                    pred_box[i].append(predictions[i]['boxes'][0])
                for j in range(1,out.shape[0]):
                    ym1 = np.where(out[j]> threshold, j+1, 0)
                    ymask += ym1
                    if predictions[i]['scores'][j] > box_threshold:
                        pred_box[i].append(predictions[i]['boxes'][j])
            else:
                ymask = np.zeros((self._data.chip_size, self._data.chip_size)) # handle for not predicted masks
            pred_mask.append(ymask)
        return pred_mask, pred_box

    def show_results(self, rows=4, mode='mask', mask_threshold=0.5, box_threshold=0.7, imsize=5, index=0, alpha=0.5, cmap='tab20', **kwargs):
        """
        Displays the results of a trained model on a part of the validation set.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        mode                    Required arguments within ['bbox', 'mask', 'bbox_mask'].
                                    * ``bbox`` - For visualizing only bounding boxes.
                                    * ``mask`` - For visualizing only mask
                                    * ``bbox_mask`` - For visualizing both mask and bounding boxes.
        ---------------------   -------------------------------------------
        mask_threshold          Optional float. The probability above which
                                a pixel will be considered mask.
        ---------------------   -------------------------------------------
        box_threshold           Optional float. The probability above which
                                a detection will be considered valid.
        ---------------------   -------------------------------------------
        nrows                   Optional int. Number of rows of results
                                to be displayed.
        =====================   ===========================================
        """
        self._check_requisites()
        if mode not in ['bbox', 'mask', 'bbox_mask']:
            raise Exception("mode can be only ['bbox', 'mask', 'bbox_mask']")

        # Get Number of items
        nrows = rows
        ncols=2


        type_data_loader = kwargs.get('data_loader', 'validation') # options : traininig, validation, testing
        if type_data_loader == 'training':
            data_loader = self._data.train_dl
        elif type_data_loader == 'validation':
            data_loader = self._data.valid_dl
        elif type_data_loader == 'testing':
            data_loader = self._data.test_dl
        else:
            e = Exception(f'could not find {type_data_loader} in data. Please ensure that the data loader type is traininig, validation or testing ')
            raise(e)

        statistics_type = kwargs.get('statistics_type', 'dataset') # Accepted Values `dataset`, `DRA`
        stretch_type = kwargs.get('stretch_type', 'minmax') # Accepted Values `minmax`, `percentclip`

        cmap_fn = getattr(matplotlib.cm, cmap)
        return_fig = kwargs.get('return_fig', False)


        x_batch, y_batch = get_nbatches(data_loader, nrows)
        x_batch = torch.cat(x_batch)
        y_batch = torch.cat(y_batch)

        nrows = min(nrows, len(x_batch))
        
        title_font_size = 16
        if kwargs.get('top', None) is not None:
            top = kwargs.get('top')
        else:
            top = 1 - (math.sqrt(title_font_size)/math.sqrt(100*nrows*imsize))



        # Get Predictions
        prediction_store = []
        for i in range(0, x_batch.shape[0], self._data.batch_size):
            prediction_store.extend(self._predict_results(x_batch[i:i+self._data.batch_size]))
        pred_mask, pred_box = self._predict_postprocess(prediction_store, mask_threshold, box_threshold)

        if self._is_multispectral:
            rgb_bands = kwargs.get('rgb_bands', self._data._symbology_rgb_bands)

            e = Exception('`rgb_bands` should be a valid band_order, list or tuple of length 3 or 1.')
            symbology_bands = []
            if not ( len(rgb_bands) == 3 or len(rgb_bands) == 1 ):
                raise(e)
            for b in rgb_bands:
                if type(b) == str:
                    b_index = self._bands.index(b)
                elif type(b) == int:
                    self._bands[b] # To check if the band index specified by the user really exists.
                    b_index = b
                else:
                    raise(e)
                b_index = self._data._extract_bands.index(b_index)
                symbology_bands.append(b_index)

             # Denormalize X
            if self._data._do_normalize:
                x_batch = (self._data._scaled_std_values[self._data._extract_bands].view(1, -1, 1, 1).to(x_batch) * x_batch ) + self._data._scaled_mean_values[self._data._extract_bands].view(1, -1, 1, 1).to(x_batch)

            # Extract RGB Bands
            symbology_x_batch = x_batch[:, symbology_bands]
            if stretch_type is not None:
                symbology_x_batch = image_batch_stretcher(symbology_x_batch, stretch_type, statistics_type)

            # Channel first to channel last for plotting
            symbology_x_batch = symbology_x_batch.permute(0, 2, 3, 1)
            # Clamp float values to range 0 - 1
            if symbology_x_batch.mean() < 1:
                symbology_x_batch = symbology_x_batch.clamp(0, 1)
        else:
            symbology_x_batch = x_batch.permute(0, 2, 3, 1)

        # Squeeze channels if single channel (1, 224, 224) -> (224, 224)
        if symbology_x_batch.shape[-1] == 1:
            symbology_x_batch = symbology_x_batch.squeeze()

        fig, ax = plt.subplots(nrows=nrows, ncols=ncols, figsize=(ncols*imsize, nrows*imsize))
        fig.suptitle('Ground Truth / Predictions', fontsize=title_font_size)
        for i in range(nrows):
            if nrows == 1:
                ax_i = ax
            else:
                ax_i = ax[i]

            # Ground Truth
            ax_i[0].imshow(symbology_x_batch[i].cpu())
            ax_i[0].axis('off')
            if mode in ['mask', 'bbox_mask']:
                n_instance = y_batch[i].unique().shape[0]
                y_merged = y_batch[i].max(dim=0)[0].cpu().numpy()
                y_rgba = cmap_fn._resample(n_instance)(y_merged)
                y_rgba[y_merged == 0] = 0
                y_rgba[:, :, -1] = alpha
                ax_i[0].imshow(y_rgba)
            ax_i[0].axis('off')

            # Predictions
            ax_i[1].imshow(symbology_x_batch[i].cpu())
            ax_i[1].axis('off')
            if mode in ['mask', 'bbox_mask']:
                n_instance = np.unique(pred_mask[i]).shape[0]
                p_rgba = cmap_fn._resample(n_instance)(pred_mask[i])
                p_rgba[pred_mask[i] == 0] = 0
                p_rgba[:, :, -1] = alpha
                ax_i[1].imshow(p_rgba)
            if mode in ['bbox_mask','bbox']:
                if pred_box[i] != []:
                    for num_boxes in pred_box[i]:
                        rect = patches.Rectangle((num_boxes[0], num_boxes[1]), num_boxes[2]-num_boxes[0], num_boxes[3]-num_boxes[1], linewidth=1, edgecolor='r', facecolor='none')
                        ax_i[1].add_patch(rect)
            ax_i[1].axis('off')
        plt.subplots_adjust(top=top)
        if self._device == torch.device('cuda'):
            torch.cuda.empty_cache()

        if return_fig:
            return fig

    def average_precision_score(self, detect_thresh=0.5, iou_thresh=0.5, mean=False, show_progress=True):

        """
        Computes average precision on the validation set for each class.

        =====================   ===========================================
        **Argument**            **Description**
        ---------------------   -------------------------------------------
        detect_thresh           Optional float. The probability above which
                                a detection will be considered for computing
                                average precision.
        ---------------------   -------------------------------------------                        
        iou_thresh              Optional float. The intersection over union
                                threshold with the ground truth mask, above
                                which a predicted mask will be
                                considered a true positive.
        ---------------------   -------------------------------------------
        mean                    Optional bool. If False returns class-wise
                                average precision otherwise returns mean
                                average precision.
        =====================   ===========================================
        :returns: `dict` if mean is False otherwise `float`
        """
        self._check_requisites()
        if mean:
            aps = compute_class_AP(self, self._data.valid_dl, 1, show_progress, detect_thresh, iou_thresh, mean)
            return aps
        else:
            aps = compute_class_AP(self, self._data.valid_dl, self._data.c - 1, show_progress, detect_thresh, iou_thresh)
            return dict(zip(self._data.classes[1:], aps))
